import time

from urbanaccess.utils import log
from urbanaccess.network import ua_network


def create_osm_net(osm_edges, osm_nodes,
                   travel_speed_mph=3, network_type='walk'):
    """
    Create a travel time weight network graph in units of minutes from
    openstreetmap nodes and edges

    Parameters
    ----------
    osm_edges : pandas.DataFrame
        osm edge dataframe
    osm_nodes : pandas.DataFrame
        osm node dataframe
    travel_speed_mph : int, optional
        travel speed to use to calculate travel time across a
        distance on a edge. units are in miles per hour (MPH)
        for pedestrian travel this is assumed to be 3 MPH
    network_type : str, optional
        default is 'walk' for the osm pedestrian network.
        this string is used to label the osm network once it is
        integrated with the transit network

    Returns
    -------
    ua_network : object
        urbanaccess_network object with osm_edges and osm_nodes dataframes
    ua_network.osm_edges : pandas.DataFrame
    ua_network.osm_nodes : pandas.DataFrame

    """
    start_time = time.time()

    if not isinstance(network_type, str) or network_type is None:
        raise ValueError('{!s} network_type passed is either not a '
                         'string or is None'.format(network_type))

    # assign impedance to OSM edges
    osm_edges['weight'] = (osm_edges[
                               'distance'] / 1609.34) / travel_speed_mph * 60

    # assign node and edge net type
    osm_edges['net_type'] = network_type
    osm_nodes['net_type'] = network_type

    ua_network.osm_nodes = osm_nodes
    ua_network.osm_edges = osm_edges

    log(
        'Created OSM network with travel time impedance using a travel speed '
        'of {} MPH. Took {:,.2f} seconds'.format(
            travel_speed_mph, time.time() - start_time))

    return ua_network



import os
import time
import pandas as pd
from re import sub
import logging as lg

from urbanaccess.utils import log


def _read_gtfs_agency(textfile_path, textfile):
    """
    Read gtfs agency.txt as a pandas dataframe

    Parameters
    ----------
    textfile_path : str
        director of text file
    textfile : str
        name of text file

    Returns
    -------
    df : pandas.DataFrame
    """
    if textfile != 'agency.txt':
        raise ValueError('{} is not a proper GTFS file name'.format(textfile))

    df = pd.read_csv(os.path.join(textfile_path, textfile), low_memory=False)
    if len(df) == 0:
        raise ValueError('{} has no records'.format(os.path.join(
            textfile_path, textfile)))
    # remove any extra whitespace in column names
    df.rename(columns=lambda x: x.strip(), inplace=True)
    return df


def _read_gtfs_stops(textfile_path, textfile):
    """
    Read gtfs stops.txt as a pandas dataframe

    Parameters
    ----------
    textfile_path : str
        director of text file
    textfile : str
        name of text file

    Returns
    -------
    df : pandas.DataFrame
    """
    if textfile != 'stops.txt':
        raise ValueError('{} is not a proper GTFS file name'.format(textfile))

    df = pd.read_csv(os.path.join(textfile_path, textfile),
                     dtype={'stop_id': object}, low_memory=False)

    if len(df) == 0:
        raise ValueError('{} has no records'.format(os.path.join(
            textfile_path, textfile)))

    df['stop_lat'] = pd.to_numeric(df['stop_lat'])
    df['stop_lon'] = pd.to_numeric(df['stop_lon'])

    # remove any extra whitespace in column names
    df.rename(columns=lambda x: x.strip(), inplace=True)
    return df


def _read_gtfs_routes(textfile_path, textfile):
    """
    Read gtfs routes.txt as a pandas dataframe

    Parameters
    ----------
    textfile_path : str
        director of text file
    textfile : str
        name of text file

    Returns
    -------
    df : pandas.DataFrame
    """
    if textfile != 'routes.txt':
        raise ValueError('{} is not a proper GTFS file name'.format(textfile))

    df = pd.read_csv(os.path.join(textfile_path, textfile),
                     dtype={'route_id': object}, low_memory=False)
    if len(df) == 0:
        raise ValueError('{} has no records'.format(os.path.join(
            textfile_path, textfile)))
    # remove any extra whitespace in column names
    df.rename(columns=lambda x: x.strip(), inplace=True)
    return df


def _read_gtfs_trips(textfile_path, textfile):
    """
    Read gtfs trips.txt as a pandas dataframe

    Parameters
    ----------
    textfile_path : str
        director of text file
    textfile : str
        name of text file

    Returns
    -------
    df : pandas.DataFrame
    """
    if textfile != 'trips.txt':
        raise ValueError('{} is not a proper GTFS file name'.format(textfile))

    df = pd.read_csv(os.path.join(textfile_path, textfile),
                     dtype={'trip_id': object,
                            'service_id': object,
                            'route_id': object,
                            7: object}, low_memory=False)
    # 7 is placeholder for shape id
    # which may not exist in some txt files
    if len(df) == 0:
        raise ValueError('{} has no records'.format(os.path.join(
            textfile_path, textfile)))
    # remove any extra whitespace in column names
    df.rename(columns=lambda x: x.strip(), inplace=True)
    return df


def _read_gtfs_stop_times(textfile_path, textfile):
    """
    Read stop_times.txt as a pandas dataframe

    Parameters
    ----------
    textfile_path : str
        director of text file
    textfile : str
        name of text file

    Returns
    -------
    df : pandas.DataFrame
    """
    if textfile != 'stop_times.txt':
        raise ValueError('{} is not a proper GTFS file name'.format(textfile))

    df = pd.read_csv(os.path.join(textfile_path, textfile),
                     dtype={'trip_id': object,
                            'stop_id': object,
                            'departure_time': object,
                            'arrival_time': object}, low_memory=False)
    if len(df) == 0:
        raise ValueError('{} has no records'.format(os.path.join(
            textfile_path, textfile)))
    # remove any extra whitespace in column names
    df.rename(columns=lambda x: x.strip(), inplace=True)
    return df


def _read_gtfs_calendar(textfile_path, textfile):
    """
    Read gtfs calendar.txt as a pandas dataframe

    Parameters
    ----------
    textfile_path : str
        director of text file
    textfile : str
        name of text file

    Returns
    -------
    df : pandas.DataFrame
    """
    if textfile != 'calendar.txt':
        raise ValueError('{} is not a proper GTFS file name'.format(textfile))

    df = pd.read_csv(os.path.join(textfile_path, textfile),
                     dtype={'service_id': object}, low_memory=False)

    if len(df) == 0:
        error_msg = ('{} has no records. This could indicate that this feed '
                     'is using calendar_dates.txt for service_ids. If so, '
                     'make a dummy row in calendar.txt to proceed.')
        raise ValueError(error_msg.format(os.path.join(textfile_path,
                                                       textfile)))

    columnlist = ['monday', 'tuesday', 'wednesday', 'thursday', 'friday',
                  'saturday', 'sunday']
    for col in columnlist:
        df[col] = pd.to_numeric(df[col])
    # remove any extra whitespace in column names
    df.rename(columns=lambda x: x.strip(), inplace=True)
    return df


def _read_gtfs_calendar_dates(textfile_path, textfile):
    """
    Read gtfs calendar_dates.txt as a pandas dataframe

    Parameters
    ----------
    textfile_path : str
        director of text file
    textfile : str
        name of text file

    Returns
    -------
    df : pandas.DataFrame
    """
    if textfile != 'calendar_dates.txt':
        raise ValueError('{} is not a proper GTFS file name'.format(textfile))

    df = pd.read_csv(os.path.join(textfile_path, textfile),
                     dtype={'service_id': object}, low_memory=False)
    if len(df) == 0:
        raise ValueError('{} has no records'.format(os.path.join(
            textfile_path, textfile)))
    # remove any extra whitespace in column names
    df.rename(columns=lambda x: x.strip(), inplace=True)
    return df


def _calendar_dates_agencyid(calendar_dates_df, routes_df,
                             trips_df, agency_df, feed_folder):
    """
    Assign unique agency id to calendar dates dataframe

    Parameters
    ----------
    calendar_dates_df : pandas:DataFrame
        calendar dates dataframe
    routes_df : pandas:DataFrame
        routes dataframe
    trips_df : pandas:DataFrame
        trips dataframe
    agency_df : pandas:DataFrame
        agency dataframe
    feed_folder : str
        name of gtfs feed folder

    Returns
    -------
    merged_df : pandas.DataFrame
    """
    tmp1 = pd.merge(routes_df, agency_df, how='left', on='agency_id',
                    sort=False, copy=False)
    tmp2 = pd.merge(trips_df, tmp1, how='left', on='route_id', sort=False,
                    copy=False)
    merged_df = pd.merge(calendar_dates_df, tmp2, how='left', on='service_id',
                         sort=False, copy=False)

    merged_df['unique_agency_id'] = _generate_unique_agency_id(merged_df,
                                                               'agency_name')

    group = merged_df[['service_id', 'unique_agency_id']].groupby([
        'service_id', 'unique_agency_id']).size()
    group_counts = group.reset_index(level=1)
    # check if service_ids are associated with more than one agency
    if any(group_counts.index.value_counts().values > 1):
        log('GTFS feed: {!s}, calendar_dates uses the same service_id across '
            'multiple agency_ids. This feed calendar_dates table will be '
            'modified from its original format to provide service_ids for '
            'each agency using a one to many join'.format(os.path.split(
                feed_folder)[1]))

        tmp = merged_df[
            ['service_id', 'unique_agency_id']].drop_duplicates(
            ['service_id', 'unique_agency_id'], inplace=False)
        merged_df = tmp.merge(calendar_dates_df, 'left',
                              on='service_id')

        return merged_df

    else:

        merged_df.drop_duplicates(subset='service_id', keep='first',
                                  inplace=True)
        merged_df = pd.merge(calendar_dates_df,
                             merged_df[['unique_agency_id', 'service_id']],
                             how='left',
                             on='service_id', sort=False, copy=False)

        return merged_df


def _calendar_agencyid(calendar_df, routes_df, trips_df,
                       agency_df, feed_folder):
    """
    Assign unique agency id to calendar dataframe

    Parameters
    ----------
    calendar_df : pandas:DataFrame
        calendar dataframe
    routes_df : pandas:DataFrame
        routes dataframe
    trips_df : pandas:DataFrame
        trips dataframe
    agency_df : pandas:DataFrame
        agency dataframe
    feed_folder : str
        name of gtfs feed folder

    Returns
    -------
    merged_df : pandas.DataFrame
    """
    tmp1 = pd.merge(routes_df, agency_df, how='left', on='agency_id',
                    sort=False, copy=False)
    tmp2 = pd.merge(trips_df, tmp1, how='left', on='route_id', sort=False,
                    copy=False)
    # do another merge to account for service ids that may not be utilized
    # across all GTFS files for accounting purposes so we keep those that
    # dont show up after merge
    merged_df = pd.merge(calendar_df[['service_id']], tmp2, how='left',
                         on='service_id', sort=False, copy=False)
    merged_df['unique_agency_id'] = _generate_unique_agency_id(merged_df,
                                                               'agency_name')

    group = merged_df[['service_id', 'unique_agency_id']].groupby([
        'service_id', 'unique_agency_id']).size()
    group_counts = group.reset_index(level=1)
    # check if service_ids are associated with more than one agency
    if any(group_counts.index.value_counts().values > 1):
        log('GTFS feed: {!s}, calendar uses the same service_id across '
            'multiple agency_ids. This feed calendar table will be '
            'modified from its original format to provide service_ids for '
            'each agency using a one to many join'.format(os.path.split(
                feed_folder)[1]))

        tmp = merged_df[
            ['service_id', 'unique_agency_id']].drop_duplicates(
            ['service_id', 'unique_agency_id'], inplace=False)
        merged_df = tmp.merge(calendar_df, 'left',
                              on='service_id')

        return merged_df

    else:

        merged_df.drop_duplicates(subset='service_id',
                                  keep='first',
                                  inplace=True)
        merged_df = pd.merge(calendar_df,
                             merged_df[['unique_agency_id', 'service_id']],
                             how='left',
                             on='service_id', sort=False, copy=False)

        return merged_df


def _trips_agencyid(trips_df, routes_df, agency_df):
    """
    Assign unique agency id to trips dataframe

    Parameters
    ----------
    trips_df : pandas:DataFrame
        trips dataframe
    routes_df : pandas:DataFrame
        routes dataframe
    agency_df : pandas:DataFrame
        agency dataframe

    Returns
    -------
    merged_df : pandas.DataFrame
    """
    tmp1 = pd.merge(routes_df, agency_df, how='left', on='agency_id',
                    sort=False, copy=False)
    merged_df = pd.merge(trips_df[['trip_id', 'route_id']], tmp1, how='left',
                         on='route_id', sort=False, copy=False)
    merged_df['unique_agency_id'] = _generate_unique_agency_id(merged_df,
                                                               'agency_name')
    merged_df.drop_duplicates(subset='trip_id', keep='first', inplace=True)

    merged_df = pd.merge(trips_df, merged_df[['unique_agency_id', 'trip_id']],
                         how='left', on='trip_id',
                         sort=False, copy=False)
    return merged_df


def _stops_agencyid(stops_df, trips_df, routes_df,
                    stop_times_df, agency_df, feed_folder):
    """
    Assign unique agency id to stops dataframe

    Parameters
    ----------
    stops_df : pandas:DataFrame
        stops dataframe
    trips_df : pandas:DataFrame
        trips dataframe
    routes_df : pandas:DataFrame
        routes dataframe
    stop_times_df : pandas:DataFrame
        stop times dataframe
    agency_df : pandas:DataFrame
        agency dataframe
    feed_folder : str
        name of gtfs feed folder

    Returns
    -------
    merged_df : pandas.DataFrame
    """
    tmp1 = pd.merge(routes_df, agency_df, how='left', on='agency_id',
                    sort=False, copy=False)
    tmp2 = pd.merge(trips_df, tmp1, how='left', on='route_id', sort=False,
                    copy=False)
    tmp3 = pd.merge(stop_times_df, tmp2, how='left', on='trip_id', sort=False,
                    copy=False)
    # do another merge to account for stops that may not be utilized across all
    # GTFS files for accounting purposes so we keep those that dont show up
    # after merge
    merged_df = pd.merge(stops_df[['stop_id']], tmp3, how='left', on='stop_id',
                         sort=False, copy=False)
    merged_df['unique_agency_id'] = _generate_unique_agency_id(merged_df,
                                                               'agency_name')

    group = merged_df[['stop_id', 'unique_agency_id']].groupby([
        'stop_id', 'unique_agency_id']).size()
    group_counts = group.reset_index(level=1)
    # check if stop_ids are associated with more than one agency
    if any(group_counts.index.value_counts().values > 1):
        log('GTFS feed: {!s}, stops uses the same stop_id across '
            'multiple agency_ids. This feed stops table will be '
            'modified from its original format to provide stop_ids for '
            'each agency using a one to many join'.format(os.path.split(
                feed_folder)[1]))

        tmp = merged_df[
            ['stop_id', 'unique_agency_id']].drop_duplicates(
            ['stop_id', 'unique_agency_id'], inplace=False)
        merged_df = tmp.merge(stops_df, 'left',
                              on='stop_id')

        return merged_df

    else:

        merged_df.drop_duplicates(subset='stop_id', keep='first', inplace=True)
        merged_df = pd.merge(stops_df,
                             merged_df[['unique_agency_id', 'stop_id']],
                             how='left', on='stop_id',
                             sort=False, copy=False)
        return merged_df


def _routes_agencyid(routes_df, agency_df):
    """
    Assign unique agency id to routes dataframe

    Parameters
    ----------
    routes_df : pandas:DataFrame
        routes dataframe
    agency_df : pandas:DataFrame
        agency dataframe

    Returns
    -------
    merged_df : pandas.DataFrame
    """
    merged_df = pd.merge(routes_df[['route_id', 'agency_id']], agency_df,
                         how='left', on='agency_id', sort=False, copy=False)
    merged_df['unique_agency_id'] = _generate_unique_agency_id(merged_df,
                                                               'agency_name')

    merged_df = pd.merge(routes_df,
                         merged_df[['unique_agency_id', 'route_id']],
                         how='left', on='route_id',
                         sort=False, copy=False)
    return merged_df


def _stop_times_agencyid(stop_times_df, routes_df, trips_df,
                         agency_df):
    """
    Assign unique agency id to stop times dataframe

    Parameters
    ----------
    stop_times_df : pandas:DataFrame
        stop times dataframe
    routes_df : pandas:DataFrame
        routes dataframe
    trips_df : pandas:DataFrame
        trips dataframe
    agency_df : pandas:DataFrame
        agency dataframe

    Returns
    -------
    merged_df : pandas.DataFrame
    """
    tmp1 = pd.merge(routes_df, agency_df, how='left', on='agency_id',
                    sort=False, copy=False)
    tmp2 = pd.merge(trips_df[['trip_id', 'route_id']], tmp1, how='left',
                    on='route_id', sort=False, copy=False)
    merged_df = pd.merge(stop_times_df, tmp2, how='left', on='trip_id',
                         sort=False, copy=False)
    merged_df['unique_agency_id'] = _generate_unique_agency_id(merged_df,
                                                               'agency_name')
    merged_df.drop_duplicates(subset='trip_id', keep='first', inplace=True)

    merged_df = pd.merge(stop_times_df,
                         merged_df[['unique_agency_id', 'trip_id']],
                         how='left', on='trip_id', sort=False, copy=False)

    return merged_df


def _add_unique_agencyid(agency_df, stops_df, routes_df,
                         trips_df, stop_times_df, calendar_df, feed_folder,
                         calendar_dates_df, nulls_as_folder=True):
    """
    Create a unique agency id for all gtfs feed dataframes to enable unique
    relational table keys

    Parameters
    ----------
    agency_df : pandas:DataFrame
        agency dataframe
    stops_df : pandas:DataFrame
        stops dataframe
    routes_df : pandas:DataFrame
        routes dataframe
    trips_df : pandas:DataFrame
        trips dataframe
    stop_times_df : pandas:DataFrame
        stop times dataframe
    calendar_df : pandas:DataFrame
        calendar dataframe
    feed_folder : str
        name of gtfs feed folder
    calendar_dates_df : pandas:DataFrame
        calendar dates dataframe
    nulls_as_folder : bool, optional
        if true, gtfs feeds where the agency id is null, the gtfs folder
        name will be used as the unique agency id
    Returns
    -------
    stops_df, routes_df, trips_df, stop_times_df, calendar_df,
    calendar_dates_df : pandas.DataFrame
    """
    start_time = time.time()

    df_list = [stops_df, routes_df, trips_df, stop_times_df, calendar_df]
    # if calendar_dates_df is not empty then add it to the processing list
    if calendar_dates_df.empty is False:
        df_list.extend([calendar_dates_df])

    path_absent = os.path.exists(
        os.path.join(feed_folder, 'agency.txt')) is False
    agency_absent = 'agency_id' not in agency_df.columns

    if (path_absent or agency_absent) and nulls_as_folder is True:

        for index, df in enumerate(df_list):
            unique_agency_id = _generate_unique_feed_id(feed_folder)
            df['unique_agency_id'] = unique_agency_id
            df_list[index] = df

        log('The agency.txt or agency_id column was not found. The unique '
            'agency id: {} was generated using the name of the folder '
            'containing the GTFS feed text files.'.format(
                unique_agency_id))

    elif path_absent is False and nulls_as_folder is False:
        raise ValueError(
            'No agency.txt file was found in {}. Add the missing file to '
            'folder or set nulls_as_folder to True'.format(feed_folder))

    else:
        if path_absent:
            file_path = os.path.join(feed_folder, 'agency.txt')
            raise ValueError('{} not found.'.format(file_path))
        if 'agency_name' not in agency_df.columns or 'agency_id' not in \
                agency_df.columns:
            raise ValueError('both agency_name and agency_id columns were not '
                             'found in agency.txt')
        if len(agency_df[['agency_id', 'agency_name']]) == 0:
            raise ValueError('agency.txt has no agency_id or agency_name '
                             'values')

        if len(agency_df['agency_name']) == 1:
            if agency_df['agency_name'].isnull().values:
                raise ValueError('null values in agency_name were found')

            # could be added to helper function
            # take first agency
            agency_snake_case = sub(r'\s+', '_', agency_df['agency_name'][0])
            unique_agency_id = agency_snake_case.replace('&', 'and').lower()

            for index, df in enumerate(df_list):
                df['unique_agency_id'] = unique_agency_id
                df_list[index] = df
            log(
                'The unique agency id: {} was generated using the name of '
                'the agency in the agency.txt file.'.format(
                    unique_agency_id))

        elif len(agency_df['agency_name']) > 1:
            if agency_df[['agency_id',
                          'agency_name']].isnull().values.any():
                raise ValueError(
                    'null values were found in agency_id and agency_name')

            # subset dataframes
            subset_agency_df = agency_df[['agency_id', 'agency_name']]
            subset_routes_df = routes_df[['route_id', 'agency_id']]
            subset_stop_times_df = stop_times_df[['trip_id', 'stop_id']]
            subset_trips_df = trips_df[['trip_id', 'route_id']]
            subset_trips_df_w_sid = trips_df[
                ['trip_id', 'route_id', 'service_id']]

            # if calendar_dates_df is not empty then process it
            if calendar_dates_df.empty is False:
                calendar_dates_replacement_df = _calendar_dates_agencyid(
                    calendar_dates_df=calendar_dates_df,
                    routes_df=subset_routes_df,
                    trips_df=subset_trips_df_w_sid,
                    agency_df=subset_agency_df,
                    feed_folder=feed_folder)

            calendar_replacement_df = _calendar_agencyid(
                calendar_df=calendar_df,
                routes_df=subset_routes_df,
                trips_df=subset_trips_df_w_sid,
                agency_df=subset_agency_df,
                feed_folder=feed_folder)

            trips_replacement_df = _trips_agencyid(
                trips_df=trips_df,
                routes_df=subset_routes_df,
                agency_df=subset_agency_df)

            stops_replacement_df = _stops_agencyid(
                stops_df=stops_df,
                trips_df=subset_trips_df,
                routes_df=subset_routes_df,
                stop_times_df=subset_stop_times_df,
                agency_df=subset_agency_df,
                feed_folder=feed_folder)

            routes_replacement_df = _routes_agencyid(
                routes_df=routes_df,
                agency_df=subset_agency_df)

            stop_times_replacement_df = _stop_times_agencyid(
                stop_times_df=stop_times_df,
                routes_df=subset_routes_df,
                trips_df=subset_trips_df,
                agency_df=subset_agency_df)

            # update the df_list object with these new variable overrides
            df_list = [stops_replacement_df,
                       routes_replacement_df,
                       trips_replacement_df,
                       stop_times_replacement_df,
                       calendar_replacement_df]
            # if calendar_dates_df is not empty then add it to the
            # processing list
            if calendar_dates_df.empty is False:
                df_list.extend([calendar_dates_replacement_df])

            log(
                'agency.txt agency_name column has more than one agency name '
                'listed. Unique agency id was assigned using the agency id '
                'and associated agency name.')

    for index, df in enumerate(df_list):
        if df['unique_agency_id'].isnull().values.any():

            unique_agency_id = _generate_unique_feed_id(feed_folder)

            df['unique_agency_id'].fillna(
                ''.join(['multiple_operators_', unique_agency_id]),
                inplace=True)
            log(
                'There are {:,} null values ({:,.2f}% of {:,} total) without '
                'a unique agency id. These records will be labeled as '
                'multiple_operators_ with the GTFS file folder '
                'name'.format(df['unique_agency_id'].isnull().sum(),
                              len(df),
                              round((float(df['unique_agency_id'].isnull(
                              ).sum()) / float(len(df)) * 100))))
            df_list[index] = df

    # if calendar_dates_df is empty then return the original empty df
    if calendar_dates_df.empty:
        df_list.extend([calendar_dates_df])

    log('Unique agency id operation complete. Took {:,.2f} seconds'.format(
        time.time() - start_time))
    return df_list


def _add_unique_gtfsfeed_id(stops_df, routes_df, trips_df,
                            stop_times_df, calendar_df, calendar_dates_df,
                            feed_folder, feed_number):
    """
    Create a unique GTFS feed specific id for all gtfs feed dataframes to
    enable tracking of specific feeds

    Parameters
    ----------
    stops_df : pandas:DataFrame
        stops dataframe
    routes_df : pandas:DataFrame
        routes dataframe
    trips_df : pandas:DataFrame
        trips dataframe
    stop_times_df : pandas:DataFrame
        stop times dataframe
    calendar_df : pandas:DataFrame
        calendar dataframe
    calendar_dates_df : pandas:DataFrame
        calendar dates dataframe
    feed_folder : str
        name of gtfs feed folder
    feed_number : int
        current number iteration of gtfs feed being read in root directory
    Returns
    -------
    stops_df, routes_df, trips_df, stop_times_df, calendar_df,
    calendar_dates_df : pandas.DataFrame
    """
    start_time = time.time()

    df_list = [stops_df,
               routes_df,
               trips_df,
               stop_times_df,
               calendar_df]
    # if calendar_dates_df is not empty then add it to the processing list
    if calendar_dates_df.empty is False:
        df_list.extend([calendar_dates_df])

    # standardize feed_folder name
    feed_folder = _generate_unique_feed_id(feed_folder)

    for index, df in enumerate(df_list):
        # create new unique_feed_id column based on the name of the feed folder
        df['unique_feed_id'] = '_'.join([feed_folder, str(feed_number)])
        df_list[index] = df

    # if calendar_dates_df is empty then return the original empty df
    if calendar_dates_df.empty:
        df_list.extend([calendar_dates_df])

    log('Unique GTFS feed id operation complete. Took {:,.2f} seconds'.format(
        time.time() - start_time))
    return df_list


def _timetoseconds(df, time_cols):
    """
    Convert default GTFS stop time departure and arrival times from 24 hour
    clock to seconds past midnight

    Parameters
    ----------
    df : pandas:DataFrame
        stop time dataframe
    time_cols : list
        list of columns to convert from 24 hour clock to seconds past
        midnight. Default column is 'departure_time'.

    Returns
    -------
    final_df : pandas.DataFrame
    """
    start_time = time.time()

    concat_series_df = pd.DataFrame()
    if not isinstance(time_cols, list):
        raise ValueError('{} is not a list'.format(time_cols))

    for col in time_cols:

        time_list = df[col].fillna('').tolist()
        for n, i in enumerate(time_list):
            if len(i) == 7:
                time_list[n] = '0' + i
            if (len(i) != 8) & (len(i) != 0) & (len(i) != 7):
                raise ValueError(
                    'Check formatting of value: {} as it is in the incorrect '
                    'format and should be 8 character string 00:00:00.'.format(
                        i))
        df_fixed = pd.DataFrame(time_list, columns=[col])

        df[col] = df_fixed[col]

        h = pd.to_numeric(df[col].str[0:2])
        if h.any() > 48:
            log('Warning: {} hour value is large and may be incorrect, '
                'please check this.'.format(
                    df[col].str[0:2].max()), level=lg.WARNING)
        m = pd.to_numeric(df[col].str[3:5])
        if m.any() > 60:
            log('Warning: {} minute value is large and may be incorrect, '
                'please check this.'.format(
                    df[col].str[3:5].max()), level=lg.WARNING)
        s = pd.to_numeric(df[col].str[6:8])
        if s.any() > 60:
            log('Warning: {} second value is large and may be incorrect, '
                'please check this.'.format(
                    df[col].str[6:8].max()), level=lg.WARNING)
        col_series = (h * 60 * 60) + (m * 60) + s
        series_df = col_series.to_frame(name=''.join([col, '_sec']))

        # check if times are negative if so display warning
        if series_df.values.any() < 0:
            log('Warning: Some stop times in {} column are negative. '
                'Time should be positive. Suggest checking original '
                'GTFS feed stop_time file before proceeding'.format(col),
                level=lg.WARNING)

        concat_series_df = pd.concat([concat_series_df, series_df], axis=1)

    final_df = pd.merge(df, concat_series_df, how='left', left_index=True,
                        right_index=True, sort=False, copy=False)

    log('Successfully converted {} to seconds past midnight and appended new '
        'columns to stop_times. Took {:,.2f} seconds'.format(
            time_cols, time.time() - start_time))

    return final_df


def _stops_definitions(df):
    """
    Append GTFS definitions for stop columns to stop dataframe

    Parameters
    ----------
    df : pandas:DataFrame
        stops dataframe

    Returns
    -------
    df : pandas.DataFrame
    """
    if 'location_type' in df.columns:
        stops_location_type = {0: 'stop', 1: 'station', 2: 'station entrance'}
        df['location_type_desc'] = df['location_type'].map(
            stops_location_type.get)

    if 'wheelchair_boarding' in df.columns:
        stops_wheelchair_boardings = {
            0: 'No accessibility information available for the stop',
            1: 'At least some vehicles at this stop can be boarded by a '
               'rider in a wheelchair',
            2: 'Wheelchair boarding is not possible at this stop'}
        df['wheelchair_boarding_desc'] = df['wheelchair_boarding'].map(
            stops_wheelchair_boardings.get)

    return df


def _routes_definitions(df):
    """
    Append GTFS definitions for route columns to route dataframe

    Parameters
    ----------
    df : pandas:DataFrame
        routes dataframe

    Returns
    -------
    df : pandas.DataFrame
    """
    if 'route_type' in df.columns:
        routes_route_type = {0: 'Street Level Rail: Tram Streetcar Light rail',
                             1: 'Underground rail: Subway or Metro',
                             2: 'Rail: intercity or long-distance ',
                             3: 'Bus',
                             4: 'Ferry',
                             5: 'Cable Car',
                             6: 'Gondola or Suspended cable car',
                             7: 'Steep incline: Funicular'}
        df['route_type_desc'] = df['route_type'].map(routes_route_type.get)

    return df


def _stop_times_definitions(df):
    """
    Append GTFS definitions for stop time columns to stop time dataframe

    Parameters
    ----------
    df : pandas:DataFrame
        stop time dataframe

    Returns
    -------
    df : pandas.DataFrame
    """
    if 'pickup_type' in df.columns:
        stop_times_pickup_type = {0: 'Regularly Scheduled',
                                  1: 'Not available',
                                  2: 'Phone arrangement only',
                                  3: 'Driver arrangement only'}
        df['pickup_type_desc'] = df['pickup_type'].map(
            stop_times_pickup_type.get)

    if 'drop_off_type' in df.columns:
        stop_times_drop_off_type = {0: 'Regularly Scheduled',
                                    1: 'Not available',
                                    2: 'Phone arrangement only',
                                    3: 'Driver arrangement only'}
        df['drop_off_type_desc'] = df['drop_off_type'].map(
            stop_times_drop_off_type.get)

    if 'timepoint' in df.columns:
        stop_times_timepoint = {'': 'Exact times',
                                0: 'Approximate times',
                                1: 'Exact times'}
        df['timepoint_desc'] = df['timepoint'].map(stop_times_timepoint.get)

    return df


def _trips_definitions(df):
    """
    Append GTFS definitions for trip columns to trip dataframe

    Parameters
    ----------
    df : pandas:DataFrame
        trip dataframe

    Returns
    -------
    df : pandas.DataFrame
    """
    if 'bikes_allowed' in df.columns:
        trips_bikes_allowed = {1: 'can accommodate at least one bicycle',
                               2: 'no bicycles are allowed on this trip'}
        df['bikes_allowed_desc'] = df['bikes_allowed'].map(
            trips_bikes_allowed.get)

    if 'wheelchair_accessible' in df.columns:
        trips_wheelchair_accessible = {
            1: 'can accommodate at least one rider in a wheelchair',
            2: 'no riders in wheelchairs can be accommodated on this trip'}
        df['wheelchair_accessible_desc'] = df['wheelchair_accessible'].map(
            trips_wheelchair_accessible.get)

    return df


def _add_txt_definitions(stops_df, routes_df, stop_times_df,
                         trips_df):
    """
    Append GTFS definitions to stops, routes, stop times, and trips dataframes

    Parameters
    ----------
    stops_df : pandas:DataFrame
        stops dataframe
    routes_df : pandas:DataFrame
        routes dataframe
    stop_times_df : pandas:DataFrame
        stop times dataframe
    trips_df : pandas:DataFrame
        trip dataframe

    Returns
    -------
    stops_df, routes_df, stop_times_df, trips_df : pandas.DataFrame
    """
    stops_df = _stops_definitions(df=stops_df)
    routes_df = _routes_definitions(df=routes_df)
    stop_times_df = _stop_times_definitions(df=stop_times_df)
    trips_df = _trips_definitions(df=trips_df)

    log(
        'Added descriptive definitions to stops, routes, stop_times, '
        'and trips tables')

    return stops_df, routes_df, stop_times_df, trips_df


def _append_route_type(stops_df, stop_times_df, routes_df,
                       trips_df, info_to_append):
    """
    Append GTFS route type definitions to stops and stop times dataframes

    Parameters
    ----------
    stops_df : pandas:DataFrame
        stops dataframe
    stop_times_df : pandas:DataFrame
        stop times dataframe
    routes_df : pandas:DataFrame
        routes dataframe
    trips_df : pandas:DataFrame
        trip dataframe
    info_to_append : {'route_type_to_stops', 'route_type_to_stop_times'}
        the type of information to append

    Returns
    -------
    stops_df or stop_times_df : pandas.DataFrame
    """
    valid_info_to_append = ['route_type_to_stops', 'route_type_to_stop_times']
    if info_to_append not in valid_info_to_append:
        raise ValueError('{} is not a valid parameter'.format(info_to_append))

    if info_to_append == 'route_type_to_stops':
        tmp1 = pd.merge(trips_df, routes_df, how='left', on='route_id',
                        sort=False, copy=False)
        merged_df = pd.merge(stop_times_df, tmp1, how='left', on='trip_id',
                             sort=False, copy=False)
        merged_df.drop_duplicates(subset='stop_id', keep='first', inplace=True)

        stops_df = pd.merge(stops_df, merged_df[['route_type', 'stop_id']],
                            how='left', on='stop_id', sort=False, copy=False)

        log('Appended route type to stops')

        return stops_df

    if info_to_append == 'route_type_to_stop_times':
        merged_df = pd.merge(trips_df, routes_df, how='left', on='route_id',
                             sort=False, copy=False)
        merged_df.drop_duplicates(subset='trip_id', keep='first', inplace=True)

        stop_times_df = pd.merge(stop_times_df,
                                 merged_df[['route_type', 'trip_id']],
                                 how='left', on='trip_id', sort=False,
                                 copy=False)

        log('Appended route type to stop_times')

        return stop_times_df


def _generate_unique_agency_id(df, col_name):
    """
    Generate unique agency id

    Parameters
    ----------
    df : pandas:DataFrame
    col_name : str
        typically will be 'agency_name'

    Returns
    -------
    col_snake_no_amps : str
    """

    col = df[col_name].astype(str)
    # replace all runs of spaces with a single underscore
    col_snake_case = col.str.replace(r'\s+', '_')
    # replace all ampersands
    col_snake_no_amps = col_snake_case.str.replace('&', 'and')
    return col_snake_no_amps.str.lower()


def _generate_unique_feed_id(feed_folder):
    """
    Generate unique feed id

    Parameters
    ----------
    feed_folder : str
        typically will be the path ot the feed folder

    Returns
    -------
    col_snake_no_amps : str
    """

    folder_name = os.path.split(feed_folder)[1]
    # replace all runs of spaces with a single underscore and replace all
    # ampersands
    folder_snake_case_no_amps = sub(r'\s+', '_', folder_name).replace('&',
                                                                      'and')

    return folder_snake_case_no_amps.lower()



from __future__ import division
import os
import logging as lg

from urbanaccess.utils import log


def _boundingbox_check(df, feed_folder, lat_min=None, lng_min=None,
                       lat_max=None, lng_max=None, bbox=None, remove=None,
                       verbose=None):
    """
    Check for and optionally remove stops that are found to be outside of a
    specified bounding box

    Parameters
    ----------
    df : pandas.DataFrame
        stops dataframe
    feed_folder : str
        name of originating gtfs feed folder
    lat_min : float
        southern latitude of bounding box
    lng_min : float
        eastern longitude of bounding box
    lat_max : float
        northern latitude of bounding box
    lng_max : float
        western longitude of bounding box
    bbox : tuple
        Bounding box formatted as a 4 element tuple: (lng_max, lat_min,
        lng_min, lat_max)
        example: (-122.304611, 37.798933, -122.263412, 37.822802)
        a bbox can be extracted for an area using: the CSV format bbox from
        http://boundingbox.klokantech.com/
    remove : bool
        if true stops that are outside the bbox will be removed
    verbose : bool
        if true and stops are found outside of the bbox, the stops that are
        outside will be printed for your reference

    Returns
    -------
    df : pandas.DataFrame

    """
    if not isinstance(verbose, bool):
        raise ValueError('verbose must be bool')
    if not isinstance(remove, bool):
        raise ValueError('remove must be bool')

    if bbox is not None:
        if not isinstance(bbox, tuple) or len(bbox) != 4:
            raise ValueError('bbox must be a 4 element tuple')
        if (lat_min is not None) or (lng_min is not None) or (
                    lat_max is not None) or (lng_max is not None):
            raise ValueError('lat_min, lng_min, lat_max and lng_max must be '
                             'None if you are using bbox')

        lng_max, lat_min, lng_min, lat_max = bbox

    if lat_min is None:
        raise ValueError('lat_min cannot be None')
    if lng_min is None:
        raise ValueError('lng_min cannot be None')
    if lat_max is None:
        raise ValueError('lat_max cannot be None')
    if lng_max is None:
        raise ValueError('lng_max cannot be None')
    if not isinstance(lat_min, float) or not isinstance(lng_min, float) or\
            not isinstance(lat_max, float) or not isinstance(lng_max, float):
        raise ValueError('lng_min, lat_min, lng_min, lat_max, and lng_max '
                         'must be floats')

    outside_boundingbox = df.loc[~(
        ((lng_max < df["stop_lon"]) & (df["stop_lon"] < lng_min)) & (
            (lat_min < df["stop_lat"]) & (df["stop_lat"] < lat_max)))]

    if len(outside_boundingbox) > 0:
        log(
            'WARNING: {} GTFS feed stops: {:,} of {:,} ({:.2f} percent of '
            'total) record(s) are outside the bounding box coordinates'.format(
                os.path.split(feed_folder)[1], len(outside_boundingbox),
                len(df), (len(outside_boundingbox) / len(df)) * 100),
            level=lg.WARNING)
        if verbose:
            log('Records: {}')
            log('{}'.format(outside_boundingbox))
        if remove:
            df_subset = df.drop(outside_boundingbox.index)
            log('Removed identified stops that are outside of bounding box.')
            return df_subset
    else:
        log(
            'No GTFS feed stops were found to be outside the bounding box '
            'coordinates')
        return df


def _checkcoordinates(df, feed_folder):
    """
    Check and print the hemisphere that stop coordinates are in

    Parameters
    ----------
    df : pandas.DataFrame
        stops dataframe
    feed_folder : str
        name of originating gtfs feed folder

    Returns
    -------
    None

    """
    if (df['stop_lat'] > 0).values.any() & (df['stop_lon'] < 0).values.any():
        log('{} GTFS feed stops: coordinates are in northwest hemisphere. '
            'Latitude = North (90); Longitude = West (-90).'.format(
                os.path.split(feed_folder)[1]))

    if (df['stop_lat'] < 0).values.any() & (df['stop_lon'] < 0).values.any():
        log('{} GTFS feed stops: coordinates are in southwest hemisphere. '
            'Latitude = South (-90); Longitude = West (-90).'.format(
                os.path.split(feed_folder)[1]))

    if (df['stop_lat'] > 0).values.any() & (df['stop_lon'] > 0).values.any():
        log('{} GTFS feed stops: coordinates are in northeast hemisphere. '
            'Latitude = North (90); Longitude = East (90).'.format(
                os.path.split(feed_folder)[1]))

    if (df['stop_lat'] < 0).values.any() & (df['stop_lon'] > 0).values.any():
        log('{} GTFS feed stops: coordinates are in southeast hemisphere. '
            'Latitude = South (-90); Longitude = East (90).'.format(
                os.path.split(feed_folder)[1]))


def _validate_gtfs(stops_df, feed_folder,
                   verbose, bbox, remove_stops_outsidebbox):
    """
    Run validation checks on stops checking for stops outside of a bounding
    box and stop coordinate hemisphere

    Parameters
    ----------
    stops_df : pandas.DataFrame
        stop times dataframe
    feed_folder : str
        name of originating gtfs feed folder
    verbose : bool
        if true and stops are found outside of the bbox, the stops that are
        outside will be printed for your reference
    bbox : tuple
        Bounding box formatted as a 4 element tuple: (lng_max, lat_min,
        lng_min, lat_max)
        example: (-122.304611, 37.798933, -122.263412, 37.822802)
        a bbox can be extracted for an area using: the CSV format bbox from
        http://boundingbox.klokantech.com/
    remove_stops_outsidebbox : bool
        if true stops that are outside the bbox will be removed

    Returns
    -------
    stops_df : pandas.DataFrame
    """

    stops_df = _boundingbox_check(df=stops_df,
                                  feed_folder=feed_folder,
                                  lat_min=None,
                                  lng_min=None,
                                  lat_max=None,
                                  lng_max=None,
                                  bbox=bbox,
                                  remove=remove_stops_outsidebbox,
                                  verbose=verbose)

    _checkcoordinates(df=stops_df, feed_folder=feed_folder)

    return stops_df



import os
import yaml


def _format_check(settings):
    """
    Check the format of a urbanaccess_config object.

    Parameters
    ----------
    settings : dict
        urbanaccess_config as a dictionary
    Returns
    -------
    Nothing
    """

    valid_keys = ['data_folder', 'logs_folder', 'log_file',
                  'log_console', 'log_name', 'log_filename', 'gtfs_api']

    for key in settings.keys():
        if key not in valid_keys:
            raise ValueError('{} not found in list of valid configuation '
                             'keys'.format(key))
        if not isinstance(key, str):
            raise ValueError('{} must be a string'.format(key))
        if key == 'log_file' or key == 'log_console':
            if not isinstance(settings[key], bool):
                raise ValueError('{} must be boolean'.format(key))


# TODO: make class CamelCase
class urbanaccess_config(object):
    """
    A set of configuration variables to initiate the
    configuration settings for UrbanAccess.

    Parameters
    ----------
    data_folder : str
        location to save and load data files
    logs_folder : str
        location to write log files
    log_file : bool
        if true, save log output to a log file in logs_folder
    log_console : bool
        if true, print log output to the console
    log_name : str
        name of the logger
    log_filename : str
        name of the log file
    gtfs_api : dict
        dictionary of the name of the GTFS API service as the key and
        the GTFS API server root URL as the value to pass to the GTFS loader
    """

    def __init__(self,
                 data_folder='data',
                 logs_folder='logs',
                 log_file=True,
                 log_console=False,
                 log_name='urbanaccess',
                 log_filename='urbanaccess',
                 gtfs_api={'gtfsdataexch': (
                         'http://www.gtfs-data-exchange.com/'
                         'api/agencies?format=csv')}):

        self.data_folder = data_folder
        self.logs_folder = logs_folder
        self.log_file = log_file
        self.log_console = log_console
        self.log_name = log_name
        self.log_filename = log_filename
        self.gtfs_api = gtfs_api

    @classmethod
    def from_yaml(cls, configdir='configs',
                  yamlname='urbanaccess_config.yaml'):
        """
        Create a urbanaccess_config instance from a saved YAML configuration.

        Parameters
        ----------
        configdir : str, optional
            Directory to load a YAML file.
        yamlname : str or file like, optional
            File name from which to load a YAML file.
        Returns
        -------
        urbanaccess_config
        """

        if not isinstance(configdir, str):
            raise ValueError('configdir must be a string')
        if not os.path.exists(configdir):
            raise ValueError('{} does not exist or was not found'.format(
                configdir))
        if not isinstance(yamlname, str):
            raise ValueError('yaml must be a string')

        yaml_file = os.path.join(configdir, yamlname)

        with open(yaml_file, 'r') as f:
            yaml_config = yaml.load(f)

        settings = cls(data_folder=yaml_config.get('data_folder', 'data'),
                       logs_folder=yaml_config.get('logs_folder', 'logs'),
                       log_file=yaml_config.get('log_file', True),
                       log_console=yaml_config.get('log_console', False),
                       log_name=yaml_config.get('log_name', 'urbanaccess'),
                       log_filename=yaml_config.get('log_filename',
                                                    'urbanaccess'),
                       gtfs_api=yaml_config.get('gtfs_api', {
                           'gtfsdataexch':
                               ('http://www.gtfs-data-exchange.com/'
                                'api/agencies?format=csv')}),
                       )

        return settings

    def to_dict(self):
        """
        Return a dict representation of an urbanaccess_config instance.
        """
        return {'data_folder': self.data_folder,
                'logs_folder': self.logs_folder,
                'log_file': self.log_file,
                'log_console': self.log_console,
                'log_name': self.log_name,
                'log_filename': self.log_filename,
                'gtfs_api': self.gtfs_api,
                }

    def to_yaml(self, configdir='configs', yamlname='urbanaccess_config.yaml',
                overwrite=False):
        """
        Save a urbanaccess_config representation to a YAML file.

        Parameters
        ----------
        configdir : str, optional
            Directory to save a YAML file.
        yamlname : str or file like, optional
            File name to which to save a YAML file.
        overwrite : bool, optional
            if true, overwrite an existing same name YAML
            file in specified directory
        Returns
        -------
        Nothing
        """

        if not isinstance(configdir, str):
            raise ValueError('configdir must be a string')
        if not os.path.exists(configdir):
            raise ValueError('{} does not exist or was not found'.format(
                configdir))
            os.makedirs(configdir)
        if not isinstance(yamlname, str):
            raise ValueError('yaml must be a string')
        yaml_file = os.path.join(configdir, yamlname)
        if overwrite is False and os.path.isfile(yaml_file) is True:
            raise ValueError(
                '{} already exists. Rename or turn overwrite to True'.format(
                    yamlname))
        else:
            with open(yaml_file, 'w') as f:
                yaml.dump(self.to_dict(), f, default_flow_style=False)


# instantiate the UrbanAccess configuration object and check format
settings = urbanaccess_config()
_format_check(settings.to_dict())



from __future__ import division
import time
from osmnet.load import network_from_bbox

import pandana
from pandana import Network

from urbanaccess.utils import log

# set the number of Pandana Networks in memory to arbitrary 40 for
# removing low connectivity nodes
# Note: reserve_num_graphs was removed in Pandana beginning in v0.4 but is
# required for prior versions
if pandana.__version__[:3] in ['0.1', '0.2', '0.3']:
    pandana.network.reserve_num_graphs(40)


def ua_network_from_bbox(lat_min=None, lng_min=None, lat_max=None,
                         lng_max=None, bbox=None, network_type='walk',
                         timeout=180, memory=None,
                         max_query_area_size=50 * 1000 * 50 * 1000,
                         remove_lcn=True):
    """
    Make a graph network (nodes and edges) from a bounding lat/lon box that
    is compatible with the network analysis tool Pandana

    Parameters
    ----------
    lat_min : float
        southern latitude of bounding box
    lng_min : float
        eastern longitude of bounding box
    lat_max : float
        northern latitude of bounding box
    lng_max : float
        western longitude of bounding box
    bbox : tuple
        Bounding box formatted as a 4 element tuple:
        (lng_max, lat_min, lng_min, lat_max)
        example: (-122.304611,37.798933,-122.263412,37.822802)
        a bbox can be extracted for an area using: the CSV format
        bbox from http://boundingbox.klokantech.com/
    network_type : {'walk', 'drive'}, optional
        Specify the network type where value of 'walk' includes roadways
        where pedestrians are allowed and pedestrian
        pathways and 'drive' includes driveable roadways. Default is walk.
    timeout : int, optional
        the timeout interval for requests and to pass to Overpass API
    memory : int, optional
        server memory allocation size for the query, in bytes. If none,
        server will use its default allocation size
    max_query_area_size : float, optional
        max area for any part of the geometry, in the units the geometry
        is in: any polygon bigger will get divided up
        for multiple queries to Overpass API
        (default is 50,000 * 50,000 units
        (ie, 50km x 50km in area, if units are meters))
    remove_lcn : bool, optional
        remove low connectivity nodes from the resulting network.
        this ensures the resulting network does
        not have nodes that are unconnected
        from the rest of the larger network

    Returns
    -------
    nodesfinal, edgesfinal : pandas.DataFrame

    """

    start_time = time.time()

    # returned osm data allows travel in both directions
    # so that all edges in integrated network are all one way edges
    two_way = False

    nodes, edges = network_from_bbox(lat_min=lat_min, lng_min=lng_min,
                                     lat_max=lat_max, lng_max=lng_max,
                                     bbox=bbox, network_type=network_type,
                                     two_way=two_way, timeout=timeout,
                                     memory=memory,
                                     max_query_area_size=max_query_area_size)

    # remove low connectivity nodes and return cleaned nodes and edges
    if remove_lcn:
        log('checking for low connectivity nodes...')
        pandana_net = Network(nodes['x'], nodes['y'],
                              edges['from'], edges['to'], edges[['distance']])
        lcn = pandana_net.low_connectivity_nodes(impedance=10000, count=10,
                                                 imp_name='distance')
        log(
            '{:,} out of {:,} nodes ({:.2f} percent of total) were '
            'identified as having low connectivity and have '
            'been removed.'.format(len(lcn), len(nodes),
                                   (len(lcn) / len(nodes)) * 100))

        rm_nodes = set(lcn)

        nodes_to_keep = ~nodes.index.isin(rm_nodes)
        edges_to_keep = ~(
            edges['from'].isin(rm_nodes) | edges['to'].isin(rm_nodes))

        nodes = nodes.loc[nodes_to_keep]
        edges = edges.loc[edges_to_keep]

        log('Completed OSM data download and graph node and edge table '
            'creation in {:,.2f} seconds'.format(time.time() - start_time))

        return nodes, edges

    else:

        log('Completed OSM data download and graph node and edge table '
            'creation in {:,.2f} seconds'.format(time.time() - start_time))

        return nodes, edges



import warnings
import pandas as pd
import time
import logging as lg

from urbanaccess.utils import log
from urbanaccess.gtfs.network import _time_selector

warnings.simplefilter(action="ignore", category=FutureWarning)


def _calc_headways_by_route_stop(df):
    """
    Calculate headways by route stop

    Parameters
    ----------
    df : pandas.DataFrame
        interpolated stop times dataframe for stop times within the time
        range with appended trip and route information

    Returns
    -------
    dataframe : pandas.DataFrame
        dataframe of statistics of route stop headways in units of minutes
    """

    # TODO: Optimize for speed

    start_time = time.time()

    df['unique_stop_route'] = (
        df['unique_stop_id'].str.cat(
            df['unique_route_id'].astype('str'), sep=','))

    stop_route_groups = df.groupby('unique_stop_route')
    log('Starting route stop headway calculation for {:,} route '
        'stops...'.format(len(stop_route_groups)))

    results = {}

    # suppress RuntimeWarning: Mean of empty slice. for this code block
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category='RuntimeWarning')

        for unique_stop_route, stop_route_group in stop_route_groups:
            stop_route_group.sort_values(['departure_time_sec_interpolate'],
                                         ascending=True, inplace=True)
            next_bus_time = (stop_route_group['departure_time_sec_interpolate']
                             .iloc[1:].values)
            prev_bus_time = (stop_route_group['departure_time_sec_interpolate']
                             .iloc[:-1].values)
            stop_route_group_headways = (next_bus_time - prev_bus_time) / 60
            results[unique_stop_route] = (pd.Series(stop_route_group_headways)
                                          .describe())

    log('Route stop headway calculation complete. Took {:,.2f} seconds'.format(
        time.time() - start_time))

    return pd.DataFrame(results).T


def _headway_handler(interpolated_stop_times_df, trips_df,
                     routes_df, headway_timerange):
    """
    route stop headway calculator handler

    Parameters
    ----------
    interpolated_stop_times_df : pandas.DataFrame
        interpolated stop times dataframe for stop times within the time range
    trips_df : pandas.DataFrame
        trips dataframe
    routes_df : pandas.DataFrame
        routes dataframe
    headway_timerange : list
        time range for which to calculate headways between as a
        list of time 1 and time 2 where times are 24 hour clock strings
        such as:
        ['07:00:00', '10:00:00']

    Returns
    -------
    headway_by_routestop_df : pandas.DataFrame
        dataframe of statistics of route stop headways in units of minutes
        with relevant route and stop information
    """
    start_time = time.time()

    # add unique trip and route id
    trips_df['unique_trip_id'] = (
        trips_df['trip_id'].str.cat(
            trips_df['unique_agency_id'].astype('str'), sep='_'))
    trips_df['unique_route_id'] = (
        trips_df['route_id'].str.cat(
            trips_df['unique_agency_id'].astype('str'), sep='_'))

    columns = ['unique_route_id', 'service_id', 'unique_trip_id',
               'unique_agency_id']
    # if these optional cols exist then keep those that do
    optional_cols = ['direction_id', 'shape_id']
    for item in optional_cols:
        if item in trips_df.columns:
            columns.append(item)

    trips_df = trips_df[columns]

    # add unique route id
    routes_df['unique_route_id'] = (
        routes_df['route_id'].str.cat(
            routes_df['unique_agency_id'].astype('str'), sep='_'))

    columns = ['unique_route_id', 'route_long_name', 'route_type',
               'unique_agency_id']
    routes_df = routes_df[columns]

    selected_interpolated_stop_times_df = _time_selector(
        df=interpolated_stop_times_df, starttime=headway_timerange[0],
        endtime=headway_timerange[1])

    tmp1 = pd.merge(trips_df, routes_df, how='left', left_on='unique_route_id',
                    right_on='unique_route_id', sort=False)
    merge_df = pd.merge(selected_interpolated_stop_times_df, tmp1, how='left',
                        left_on='unique_trip_id', right_on='unique_trip_id',
                        sort=False)
    cols_to_drop = ['unique_agency_id_y', 'unique_agency_id_x']
    merge_df.drop(cols_to_drop, axis=1, inplace=True)

    headway_by_routestop_df = _calc_headways_by_route_stop(df=merge_df)

    # add unique route stop node_id
    headway_by_routestop_df = pd.merge(
        headway_by_routestop_df,
        merge_df[['unique_stop_route', 'unique_stop_id', 'unique_route_id']],
        how='left', left_index=True, right_on='unique_stop_route', sort=False)
    headway_by_routestop_df.drop('unique_stop_route', axis=1, inplace=True)
    headway_by_routestop_df['node_id_route'] = (
        headway_by_routestop_df['unique_stop_id'].str.cat(
            headway_by_routestop_df['unique_route_id'].astype('str'), sep='_'))

    log('headway calculation complete. Took {:,.2f} seconds'.format(
        time.time() - start_time))

    return headway_by_routestop_df


def headways(gtfsfeeds_df, headway_timerange):
    """
    Calculate headways by route stop for a specific time range

    Parameters
    ----------
    gtfsfeeds_df : object
        gtfsfeeds_dfs object with all processed GTFS data tables
    headway_timerange : list
        time range for which to calculate headways between as a list of
        time 1 and time 2 where times are 24 hour clock strings such as:
        ['07:00:00', '10:00:00']

    Returns
    -------
    gtfsfeeds_dfs.headways : pandas.DataFrame
        gtfsfeeds_dfs object for the headways dataframe with statistics of
        route stop headways in units of minutes
        with relevant route and stop information
    """

    time_error_statement = (
        '{} starttime and endtime are not in the correct format. '
        'Format should be a 24 hour clock in following format: 08:00:00 '
        'or 17:00:00'.format(headway_timerange))
    if not isinstance(headway_timerange, list) or len(headway_timerange) != 2:
        raise ValueError('timerange must be a list of length 2')
    if headway_timerange[0].split(':')[0] > headway_timerange[1].split(':')[0]:
        raise ValueError('starttime is greater than endtime')

    for t in headway_timerange:
        if not isinstance(t, str):
            raise ValueError(time_error_statement)
        if len(t) != 8:
            raise ValueError(time_error_statement)
    if int(headway_timerange[1].split(':')[0]) - int(
            headway_timerange[0].split(':')[0]) > 3:
        long_time_range_msg = (
            'WARNING: Time range passed: {} is a {} hour period. Long periods '
            'over 3 hours may take a significant amount of time to process.')
        log(long_time_range_msg.format(headway_timerange,
                                       int(str(
                                           headway_timerange[1][0:2])) - int(
                                           str(headway_timerange[0][0:2]))),
            level=lg.WARNING)

    if gtfsfeeds_df is None:
        raise ValueError('gtfsfeeds_df cannot be None')
    if gtfsfeeds_df.stop_times_int.empty or gtfsfeeds_df.trips.empty or \
            gtfsfeeds_df.routes.empty:
        raise ValueError(
            'one of the gtfsfeeds_dfs objects: stop_times_int, trips, '
            'or routes were found to be empty.')

    headways_df = _headway_handler(
        interpolated_stop_times_df=gtfsfeeds_df.stop_times_int,
        trips_df=gtfsfeeds_df.trips,
        routes_df=gtfsfeeds_df.routes,
        headway_timerange=headway_timerange)

    gtfsfeeds_df.headways = headways_df

    return gtfsfeeds_df



import yaml
import pandas as pd
import traceback
import zipfile
import os
import logging as lg
import time
from six.moves.urllib.request import urlopen

from urbanaccess.utils import log
from urbanaccess import config


# TODO: make class CamelCase
class urbanaccess_gtfsfeeds(object):
    """
    A dict of GTFS feeds as {name of GTFS feed or transit service/agency :
    URL of feed} to request and
    download in the gtfs downloader.

    Parameters
    ----------
    gtfs_feeds : dict
        dictionary of the name of the transit service or agency GTFS feed
        as the key -note: this name will be used as the feed folder name.
        If the GTFS feed does not have a agency name in the agency.txt file
        this key will be used to name the agency- and
        the GTFS feed URL as the value to pass to the GTFS downloader as:
        {unique name of GTFS feed or transit service/agency : URL of feed}
    """

    def __init__(self,
                 gtfs_feeds={}):

        self.gtfs_feeds = gtfs_feeds

    @classmethod
    def from_yaml(cls, gtfsfeeddir=os.path.join(config.settings.data_folder,
                                                'gtfsfeeds'),
                  yamlname='gtfsfeeds.yaml'):
        """
        Create a urbanaccess_gtfsfeeds instance from a saved YAML.

        Parameters
        ----------
        gtfsfeeddir : str, optional
            Directory to load a YAML file.
        yamlname : str or file like, optional
            File name from which to load a YAML file.
        Returns
        -------
        urbanaccess_gtfsfeeds
        """

        if not isinstance(gtfsfeeddir, str):
            raise ValueError('gtfsfeeddir must be a string')
        if not os.path.exists(gtfsfeeddir):
            raise ValueError('{} does not exist or was not found'.format(
                gtfsfeeddir))
        if not isinstance(yamlname, str):
            raise ValueError('yaml must be a string')

        yaml_file = os.path.join(gtfsfeeddir, yamlname)

        with open(yaml_file, 'r') as f:
            yaml_config = yaml.load(f)

        if not isinstance(yaml_config, dict):
            raise ValueError('{} yamlname is not a dict'.format(yamlname))

        validkey = 'gtfs_feeds'
        if validkey not in yaml_config.keys():
            raise ValueError('key gtfs_feeds was not found in YAML file')

        for key in yaml_config['gtfs_feeds'].keys():
            if not isinstance(key, str):
                raise ValueError('{} must be a string'.format(key))
            for value in yaml_config['gtfs_feeds'][key]:
                if not isinstance(value, str):
                    raise ValueError('{} must be a string'.format(value))

        if (pd.Series(
                yaml_config['gtfs_feeds'].values()).value_counts() != 1).all():
            raise ValueError(
                'duplicate values were found when the passed add_dict '
                'dictionary was added to the existing dictionary. Feed URL '
                'values must be unique.')

        gtfsfeeds = cls(gtfs_feeds=yaml_config.get('gtfs_feeds', {}))
        log('{} YAML successfully loaded with {} feeds.'.format(yaml_file, len(
            yaml_config['gtfs_feeds'])))

        return gtfsfeeds

    def to_dict(self):
        """
        Return a dict representation of an urbanaccess_gtfsfeeds instance.
        """
        return {'gtfs_feeds': self.gtfs_feeds}

    def add_feed(self, add_dict, replace=False):
        """
        Add a dictionary to the urbanaccess_gtfsfeeds instance.

        Parameters
        ----------
        add_dict : dict
            Dictionary to add to existing urbanaccess_gtfsfeeds with the name
            of the transit service or agency GTFS feed as the key and the
            GTFS feed URL as the value to pass to the GTFS downloader
            as:
            {unique name of GTFS feed or transit service/agency : URL of feed}
        replace : bool, optional
            If key of dict is already in the UrbanAccess replace
            the existing dict value with the value passed
        """

        if not isinstance(add_dict, dict):
            raise ValueError('add_dict is not a dict')
        if not isinstance(replace, bool):
            raise ValueError('replace is not bool')

        if replace is not True:

            for key in add_dict.keys():
                if key in self.gtfs_feeds.keys():
                    raise ValueError(
                        '{} passed in add_dict already exists in gtfs_feeds. '
                        'Only unique keys are allowed to be added.'.format(
                            key))
                if not isinstance(key, str):
                    raise ValueError('{} must be a string'.format(key))
                for value in add_dict[key]:
                    if not isinstance(value, str):
                        raise ValueError('{} must be a string'.format(value))

            for key, value in add_dict.items():
                if value in self.gtfs_feeds.values():
                    raise ValueError('duplicate values were found when the '
                                     'passed add_dict dictionary was added to '
                                     'the existing dictionary. Feed URL '
                                     'values must be unique.')
            gtfs_feeds = self.gtfs_feeds.update(add_dict)

        else:
            for key in add_dict.keys():
                if key in self.gtfs_feeds.keys():
                    log('{} passed in add_dict will replace existing {} feed '
                        'in gtfs_feeds.'.format(key, key))
                if not isinstance(key, str):
                    raise ValueError('{} must be a string'.format(key))
                for value in add_dict[key]:
                    if not isinstance(value, str):
                        raise ValueError('{} must be a string'.format(value))

            gtfs_feeds = self.gtfs_feeds.update(add_dict)

        log('Added {} feeds to gtfs_feeds: {}'.format(len(add_dict), add_dict))

        return gtfs_feeds

    def remove_feed(self, del_key=None, remove_all=False):
        """
        Remove GTFS feeds from the existing urbanaccess_gtfsfeeds instance

        Parameters
        ----------
        del_key : str or list, optional
            dict keys as a single string or list of
            strings to remove from existing
        remove_all : bool, optional
            if true, remove all keys from existing
            urbanaccess_gtfsfeeds instance
        """

        if not isinstance(remove_all, bool):
            raise ValueError('remove_all is not bool')

        if del_key is None and remove_all:
            self.gtfs_feeds = {}
            log('Removed all feeds from gtfs_feeds')

        else:

            if not isinstance(del_key, (list, str)):
                raise ValueError('del_key must be a string or list of strings')
            if remove_all:
                raise ValueError(
                    'remove_all must be False in order to remove individual '
                    'records: {}'.format(del_key))

            del_key = [del_key]

            for key in del_key:
                if key not in self.gtfs_feeds.keys():
                    raise ValueError(
                        '{} key to delete was not found in gtfs_feeds'.format(
                            key))
                del self.gtfs_feeds[key]
                log('Removed {} feed from gtfs_feeds'.format(key))

    def to_yaml(self, gtfsfeeddir=os.path.join(config.settings.data_folder,
                                               'gtfsfeeds'),
                yamlname='gtfsfeeds.yaml',
                overwrite=False):
        """
        Save a urbanaccess_gtfsfeeds representation to a YAML file.

        Parameters
        ----------
        gtfsfeeddir : str, optional
            Directory to save a YAML file.
        yamlname : str or file like, optional
            File name to which to save a YAML file.
        overwrite : bool, optional
            if true, overwrite an existing same name YAML file in specified
            directory
        Returns
        -------
        Nothing
        """

        if not isinstance(gtfsfeeddir, str):
            raise ValueError('gtfsfeeddir must be a string')
        if not os.path.exists(gtfsfeeddir):
            log(
                '{} does not exist or was not found and will be '
                'created'.format(
                    gtfsfeeddir))
            os.makedirs(gtfsfeeddir)
        if not isinstance(yamlname, str):
            raise ValueError('yaml must be a string')
        yaml_file = os.path.join(gtfsfeeddir, yamlname)
        if overwrite is False and os.path.isfile(yaml_file) is True:
            raise ValueError(
                '{} already exists. Rename or turn overwrite to True'.format(
                    yamlname))
        else:
            with open(yaml_file, 'w') as f:
                yaml.dump(self.to_dict(), f, default_flow_style=False)
            log('{} file successfully created'.format(yaml_file))


# instantiate the UrbanAccess gtfs feed object
feeds = urbanaccess_gtfsfeeds()


def search(api='gtfsdataexch', search_text=None, search_field=None,
           match='contains', add_feed=False, overwrite_feed=False):
    """
    Connect to a GTFS feed repository API and search for GTFS feeds that exist
    in a remote GTFS repository and whether or not to add the GTFS feed name
    and download URL to the urbanaccess_gtfsfeeds instance.
    Currently only supports access to the GTFS Data Exchange API.

    Parameters
    ----------
    api : {'gtfsdataexch'}, optional
        name of GTFS feed repository to search in. name corresponds to the
        dict specified in the urbanacess_config instance. Currently only
        supports access to the GTFS Data Exchange repository.
    search_text : str, optional
        string pattern to search for
    search_field : string or list, optional
        name of the field or column to search for string
    match : {'contains', 'exact'}, optional
        search string matching method as either: contains or exact
    add_feed : bool, optional
        add search results to existing urbanaccess_gtfsfeeds instance using
        the name field as the key and the URL as the value
    overwrite_feed : bool, optional
        If true the existing urbanaccess_gtfsfeeds instance will be replaced
        with the records returned in the search results.
        All existing records will be removed.
    Returns
    -------
    search_result_df : pandas.DataFrame
        Dataframe of search results displaying full feed metadata
    """

    log(
        'Note: Your use of a GTFS feed is governed by each GTFS feed author '
        'license terms. It is suggested you read the respective license '
        'terms for the appropriate use of a GTFS feed.',
        level=lg.WARNING)

    if not isinstance(api, str):
        raise ValueError('{} must be a string'.format(api))
    if api not in config.settings.gtfs_api.keys():
        raise ValueError('{} is not currently a supported API'.format(api))
    if config.settings.gtfs_api[api] is None or not isinstance(
            config.settings.gtfs_api[api], str):
        raise ValueError('{} is not defined or defined incorrectly'.format(
            api))
    if not isinstance(match, str) or match not in ['contains', 'exact']:
        raise ValueError('match must be either: contains or exact')
    if not isinstance(add_feed, bool):
        raise ValueError('add_feed must be bool')

    if api == 'gtfsdataexch':
        log(
            'Warning: The GTFSDataExchange is no longer being maintained as '
            'of Summer 2016. '
            'Data accessed here may be out of date.', level=lg.WARNING)

        feed_table = pd.read_table(config.settings.gtfs_api[api], sep=',')
        feed_table['date_added'] = pd.to_datetime(feed_table['date_added'],
                                                  unit='s')
        feed_table['date_last_updated'] = pd.to_datetime(
            feed_table['date_last_updated'], unit='s')

        if search_text is None:
            log(
                'No search parameters were passed. Returning full list of {} '
                'GTFS feeds:'.format(
                    len(feed_table)))
            return feed_table
        else:
            pass

        search_result_df = pd.DataFrame()

        if search_field is None:
            search_field = ['name', 'url', 'dataexchange_id', 'feed_baseurl']
        else:
            if not isinstance(search_field, list):
                raise ValueError('search_field is not list')

        for field in search_field:
            if field not in feed_table.columns:
                raise ValueError(
                    '{} column not found in available feed table'.format(
                        field))
            for col in feed_table.select_dtypes(include=[object]).columns:

                if isinstance(search_text, str):
                    search_text = [search_text]
                else:
                    if not isinstance(search_text, list):
                        raise ValueError('search_text is not list')

                for text in search_text:
                    if match == 'contains':
                        search_result = feed_table[
                            feed_table[col].str.contains(text, case=False,
                                                         na=False)]
                    if match == 'exact':
                        search_result = feed_table[
                            feed_table[col].str.match(text, case=False,
                                                      na=False)]
                    search_result_df = search_result_df.append(search_result)
                    search_result_df.drop_duplicates(inplace=True)

        log('Found {} records that matched {} inside {} columns:'.format(
            len(search_result_df), search_text, search_field))

        if len(search_result_df) != 0:

            if add_feed:
                if overwrite_feed:
                    zip_url = search_result_df[
                                  'dataexchange_url'] + 'latest.zip'
                    search_result_df['dataexchange_url'] = zip_url
                    search_result_dict = search_result_df.set_index('name')[
                        'dataexchange_url'].to_dict()
                    feeds.gtfs_feeds = search_result_dict
                    log(
                        'Replaced all records in gtfs_feed list with the {} '
                        'found records:'.format(
                            len(search_result_df)))
                else:
                    zip_url = search_result_df[
                                  'dataexchange_url'] + 'latest.zip'
                    search_result_df['dataexchange_url'] = zip_url
                    search_result_dict = search_result_df.set_index('name')[
                        'dataexchange_url'].to_dict()
                    feeds.add_feed(search_result_dict)
                    log('Added {} records to gtfs_feed list:'.format(
                        len(search_result_df)))
                return search_result_dict

            else:
                return search_result_df


def download(data_folder=os.path.join(config.settings.data_folder),
             feed_name=None, feed_url=None, feed_dict=None,
             error_pause_duration=5, delete_zips=False):
    """
    Connect to the URLs passed in function or the URLs stored in the
    urbanaccess_gtfsfeeds instance and download the GTFS feed zipfile(s)
    then unzip inside a local root directory. Resulting GTFS feed text files
    will be located in the root folder: gtfsfeed_text unless otherwise
    specified

    Parameters
    ----------
    data_folder : str, optional
        directory to download GTFS feed data to
    feed_name : str, optional
        name of transit agency or service to use to name downloaded zipfile
    feed_url : str, optional
        corresponding URL to the feed_name to use to download GTFS feed zipfile
    feed_dict : dict, optional
        Dictionary specifying the name of the transit service or
        agency GTFS feed as the key and the GTFS feed URL as the value:
        {unique name of GTFS feed or transit service/agency : URL of feed}
    error_pause_duration : int, optional
        how long to pause in seconds before re-trying requests if error
    delete_zips : bool, optional
        if true the downloaded zipfiles will be removed
    Returns
    -------
    nothing
    """

    if (feed_name is not None and feed_url is None) or (
            feed_url is not None and feed_name is None):
        raise ValueError(
            'Both feed_name and feed_url parameters are required.')

    if feed_name is not None and feed_url is not None:
        if feed_dict is not None:
            raise ValueError('feed_dict is not specified')
        if not isinstance(feed_name, str) or not isinstance(feed_url, str):
            raise ValueError('either feed_name and or feed_url are not string')
        feeds.gtfs_feeds = {feed_name: feed_url}

    elif feed_dict is not None:
        if feed_name is not None or feed_url is not None:
            raise ValueError('either feed_name and or feed_url are not None')
        if not isinstance(feed_dict, dict):
            raise ValueError('feed_dict is not dict')
        for key in feed_dict.keys():
            if not isinstance(key, str):
                raise ValueError('{} must be a string'.format(key))
            for value in feed_dict[key]:
                if not isinstance(value, str):
                    raise ValueError('{} must be a string'.format(value))

        for key, value in feed_dict.items():
            if value in feed_dict.gtfs_feeds.values():
                raise ValueError(
                    'duplicate values were found when the passed add_dict '
                    'dictionary was added to the existing dictionary. Feed '
                    'URL values must be unique.')

        feeds.gtfs_feeds = feed_dict
    elif feed_name is None and feed_url is None and feed_dict is None:
        if len(feeds.gtfs_feeds) == 0:
            raise ValueError('No records were found in passed feed_dict')
        feeds.gtfs_feeds
    else:
        raise ValueError('Passed parameters were incorrect or not specified.')

    download_folder = os.path.join(data_folder, 'gtfsfeed_zips')

    if not os.path.exists(download_folder):
        os.makedirs(download_folder)
        log('{} does not exist. Directory was created'.format(download_folder))
    log('{} GTFS feeds will be downloaded here: {}'.format(
        len(feeds.gtfs_feeds), download_folder))

    start_time1 = time.time()
    # TODO: add file counter and print number to user
    for feed_name_key, feed_url_value in feeds.gtfs_feeds.items():
        start_time2 = time.time()
        zipfile_path = ''.join([download_folder, '/', feed_name_key, '.zip'])

        if 'http' in feed_url_value:
            status_code = urlopen(feed_url_value).getcode()
            if status_code == 200:
                file = urlopen(feed_url_value)

                _zipfile_type_check(file=file,
                                    feed_url_value=feed_url_value)

                with open(zipfile_path, "wb") as local_file:
                    local_file.write(file.read())
                log(
                    '{} GTFS feed downloaded successfully. Took {:,'
                    '.2f} seconds for {:,.1f}KB'.format(
                        feed_name_key, time.time() - start_time2,
                        os.path.getsize(zipfile_path)))
            elif status_code in [429, 504]:
                log(
                    'URL at {} returned status code {} and no data. '
                    'Re-trying request in {:.2f} seconds.'.format(
                        feed_url_value, status_code, error_pause_duration),
                    level=lg.WARNING)
                time.sleep(error_pause_duration)
                try:
                    file = urlopen(feed_url_value)

                    _zipfile_type_check(file=file,
                                        feed_url_value=feed_url_value)

                    with open(zipfile_path, "wb") as local_file:
                        local_file.write(file.read())
                except Exception:
                    log('Unable to connect. URL at {} returned status code '
                        '{} and no data'.format(feed_url_value, status_code),
                        level=lg.ERROR)
            else:
                log(
                    'Unable to connect. URL at {} returned status code {} '
                    'and no data'.format(
                        feed_url_value, status_code), level=lg.ERROR)
        else:
            try:
                file = urlopen(feed_url_value)
                _zipfile_type_check(file=file,
                                    feed_url_value=feed_url_value)
                with open(
                        ''.join([download_folder, '/', feed_name_key, '.zip']),
                        "wb") as local_file:
                    local_file.write(file.read())
                log(
                    '{} GTFS feed downloaded successfully. Took {:,'
                    '.2f} seconds for {:,.1f}KB'.format(
                        feed_name_key, time.time() - start_time2,
                        os.path.getsize(zipfile_path)))
            except Exception:
                log('Unable to connect: {}'.format(traceback.format_exc()),
                    level=lg.ERROR)

    log('GTFS feed download completed. Took {:,.2f} seconds'.format(
        time.time() - start_time1))

    _unzip(zip_rootpath=download_folder, delete_zips=delete_zips)


def _unzip(zip_rootpath, delete_zips=True):
    """
    unzip all GTFS feed zipfiles in a root directory with resulting text files
    in the root folder: gtfsfeed_text

    Parameters
    ----------
    zip_rootpath : string
        root directory to place downloaded GTFS feed zipfiles
    delete_zips : bool, optional
        if true the downloaded zipfiles will be removed

    Returns
    -------
    nothing
    """

    start_time = time.time()

    unzip_rootpath = os.path.join(os.path.dirname(zip_rootpath),
                                  'gtfsfeed_text')

    if not os.path.exists(unzip_rootpath):
        os.makedirs(unzip_rootpath)
        log('{} does not exist. Directory was created'.format(unzip_rootpath))

    zipfilelist = [zipfilename for zipfilename in os.listdir(zip_rootpath) if
                   zipfilename.endswith(".zip")]
    if len(zipfilelist) == 0:
        raise ValueError('No zipfiles were found in specified '
                         'directory: {}'.format(zip_rootpath))

    for zfile in zipfilelist:

        with zipfile.ZipFile(os.path.join(zip_rootpath, zfile)) as z:
            # required to deal with zipfiles that have subdirectories and
            # that were created on OSX
            filelist = [file for file in z.namelist() if
                        file.endswith(".txt") and not file.startswith(
                            "__MACOSX")]
            if not os.path.exists(
                    os.path.join(unzip_rootpath, zfile.replace('.zip', ''))):
                os.makedirs(
                    os.path.join(unzip_rootpath, zfile.replace('.zip', '')))
            for file in filelist:
                with open(
                        os.path.join(unzip_rootpath, zfile.replace('.zip', ''),
                                     os.path.basename(file)), 'wb') as f:
                    f.write(z.read(file))
                    f.close()
            z.close()
            log('{} successfully extracted to: {}'.format(zfile, os.path.join(
                unzip_rootpath, zfile.replace('.zip', ''))))
    if delete_zips:
        os.remove(zip_rootpath)
        log('Deleted {} folder'.format(zip_rootpath))
    log(
        'GTFS feed zipfile extraction completed. Took {:,.2f} seconds for {} '
        'files'.format(
            time.time() - start_time, len(zipfilelist)))


def _zipfile_type_check(file, feed_url_value):
    """
    zipfile format checker helper

    Parameters
    ----------
    file : addinfourl
        loaded zipfile object in memory
    feed_url_value : str
        URL to download GTFS feed zipfile

    Returns
    -------
    nothing
    """
    if 'zip' not in file.info().get('Content-Type') is True \
            or 'octet' not in file.info().get('Content-Type') is True:
        raise ValueError(
            'data requested at {} is not a zipfile. '
            'Data must be a zipfile'.format(feed_url_value))



# The following logging functions were modified from the osmnx library and
# used with permission from the author Geoff Boeing:
# log, _get_logger: https://github.com/gboeing/osmnx/blob/master/osmnx/utils.py

import logging as lg
import unicodedata
import sys
import datetime as dt
import os
import pandas as pd

from urbanaccess import config


def log(message, level=None, name=None, filename=None):
    """
    Write a message to the log file and/or print to the console.

    Parameters
    ----------
    message : string
        the content of the message to log
    level : int, optional
        one of the logger.level constants
    name : string, optional
        name of the logger
    filename : string, optional
        name of the log file

    Returns
    -------
    None
    """

    if level is None:
        level = lg.INFO
    if name is None:
        name = config.settings.log_name
    if filename is None:
        filename = config.settings.log_filename

    if config.settings.log_file:
        # get the current logger or create a new one then log message at
        # requested level
        logger = _get_logger(level=level, name=name, filename=filename)
        if level == lg.DEBUG:
            logger.debug(message)
        elif level == lg.INFO:
            logger.info(message)
        elif level == lg.WARNING:
            logger.warning(message)
        elif level == lg.ERROR:
            logger.error(message)

    # if logging to console is turned on, convert message to ascii and print
    #  to the console only
    if config.settings.log_console:
        # capture current stdout, then switch it to the console, print the
        # message, then switch back to what had been the stdout
        # this prevents logging to notebook - instead, it goes to console
        standard_out = sys.stdout
        sys.stdout = sys.__stdout__

        # convert message to ascii for proper console display in windows
        # terminals
        message = unicodedata.normalize('NFKD', unicode(message)).encode(
            'ascii', errors='replace').decode()
        print(message)
        sys.stdout = standard_out
    # otherwise print out standard statement
    else:
        print(message)


def _get_logger(level=None, name=None, filename=None):
    """
    Create a logger or return the current one if already instantiated.

    Parameters
    ----------
    level : int, optional
        one of the logger.level constants
    name : string, optional
        name of the logger
    filename : string, optional
        name of the log file

    Returns
    -------
    logger : logger.logger
    """

    if level is None:
        level = config.settings.log_level
    if name is None:
        name = config.settings.log_name
    if filename is None:
        filename = config.settings.log_filename

    logger = lg.getLogger(name)

    # if a logger with this name is not already established
    if not getattr(logger, 'handler_set', None):

        todays_date = dt.datetime.today().strftime('%Y_%m_%d')
        log_filename = '{}/{}_{}.log'.format(config.settings.logs_folder,
                                             filename, todays_date)

        if not os.path.exists(config.settings.logs_folder):
            os.makedirs(config.settings.logs_folder)

        # create file handler and log formatter and establish settings
        handler = lg.FileHandler(log_filename, encoding='utf-8')
        formatter = lg.Formatter(
            '%(asctime)s %(levelname)s %(name)s %(message)s')
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        logger.setLevel(level)
        logger.handler_set = True

    return logger


def create_hdf5(dir=None, filename=None, overwrite_hdf5=False):
    """
    Create a empty hdf5 file

    Parameters
    ----------
    dir : string, optional
        directory to save hdf5 file, if None defaults to dir set in
        config.settings.data_folder
    filename : string, optional
        name of the hdf5 file to save with .h5 extension, if None defaults
        to urbanaccess.h5
    overwrite_hdf5 : bool, optional
        if true any existing hdf5 file with the specified name in the
        specified directory will be overwritten

    Returns
    -------
    None
    """
    if dir is None:
        dir = config.settings.data_folder
    else:
        if not isinstance(dir, str):
            raise ValueError('Directory must be a string')

    try:
        if not os.path.exists(dir):
            os.makedirs(dir)
    except Exception:
        raise ValueError('Unable to make directory {}'.format(dir))

    if filename is None:
        filename = 'urbanaccess.h5'
    else:
        if not isinstance(filename, str):
            raise ValueError('Filename must be a string')

    hdf5_save_path = '{}/{}'.format(dir, filename)
    if not filename.endswith('.h5'):
        raise ValueError('hdf5 filename extension must be "h5"')

    if not os.path.exists(hdf5_save_path):
        store = pd.HDFStore(hdf5_save_path)
        store.close()
        log('New {} hdf5 store created in dir: {}'.format(filename, dir))
    elif overwrite_hdf5 and os.path.exists(hdf5_save_path):
        store = pd.HDFStore(hdf5_save_path)
        store.close()
        log('Existing {} hdf5 store in dir: has been overwritten.'.format(
            hdf5_save_path))
    else:
        log('Using existing {} hdf5 store.'.format(hdf5_save_path))

    return hdf5_save_path


def df_to_hdf5(data=None, key=None, overwrite_key=False, dir=None,
               filename=None, overwrite_hdf5=False):
    """
    Write a pandas dataframe to a table in a hdf5 file

    Parameters
    ----------
    data : pandas.DataFrame
        pandas dataframe to save to a hdf5 table
    key : string
        name of table to save dataframe as in the hdf5 file
    overwrite_key : bool, optional
        if true any existing table with the specified key name will be
        overwritten
    dir : string
        directory to save hdf5 file
    filename : string
        name of the hdf5 file to save with .h5 extension
    overwrite_hdf5 : bool, optional
        if true any existing hdf5 file with the specified name in the
        specified directory will be overwritten

    Returns
    -------
    None
    """
    hdf5_save_path = create_hdf5(dir=dir, filename=filename,
                                 overwrite_hdf5=overwrite_hdf5)

    store = pd.HDFStore(hdf5_save_path, mode='r')

    if not ''.join(['/', key]) in store.keys():
        store.close()
        data.to_hdf(hdf5_save_path, key=key, mode='a', format='table')
        log('{} saved in {} hdf5 store.'.format(key, hdf5_save_path))

    elif ''.join(['/', key]) in store.keys() and overwrite_key:
        store.close()
        data.to_hdf(hdf5_save_path, key=key, mode='a', format='table')
        log('Existing {} overwritten in {} hdf5 store.'.format(key,
                                                               hdf5_save_path))

    else:
        store.close()
        log(
            'Key {} already exists in {} hdf5 store. Set to overwrite_key = '
            'True to replace.'.format(
                key, hdf5_save_path))


def hdf5_to_df(dir=None, filename=None, key=None):
    """
    Read data from a hdf5 file to a pandas dataframe

    Parameters
    ----------
    dir : string
        directory of the hdf5 file to read from
    filename : string
        name of the hdf5 file with .h5 extension to read from
    key : string
        table inside the hdf5 file to return as a pandas dataframe

    Returns
    -------
    df : pandas.DataFrame
    """
    if dir is None:
        dir = config.settings.data_folder
    else:
        if not isinstance(dir, str):
            raise ValueError('Directory must be a string')

    if filename is None:
        filename = 'urbanaccess_net.h5'
    else:
        if not isinstance(filename, str):
            raise ValueError('Filename must be a string')

    hdf5_load_path = '{}/{}'.format(dir, filename)

    if not filename.endswith('.h5'):
        raise ValueError('hdf5 filename extension must be "h5"')
    if not os.path.exists(hdf5_load_path):
        raise ValueError('Unable to find directory or file: {}'.format(
            hdf5_load_path))

    with pd.HDFStore(hdf5_load_path) as store:
        # TODO: fix print statement to only display current key, not all keys
        log('Successfully read store: {} with the following keys: {}'.format(
            hdf5_load_path, store.keys()))
        try:
            df = store[key]
            ('Returned {} as dataframe'.format(key))
        except Exception:
            raise ValueError(
                'Unable to find key: {}. Keys found: {}'.format(key,
                                                                store.keys()))

        return df



from .gtfs.load import *
from .gtfs.network import *
from .gtfs.gtfsfeeds_dataframe import *
from .gtfs.headways import *
from .osm.load import *
from .osm.network import *
from .network import *
from .utils import *
from .gtfsfeeds import *
from .plot import *

__version__ = "0.2.0"

version = __version__



import time
from sklearn.neighbors import KDTree
import pandas as pd
from geopy.distance import vincenty

from urbanaccess.utils import log, df_to_hdf5, hdf5_to_df
from urbanaccess import config


class urbanaccess_network(object):
    """
    A urbanaccess object of Pandas DataFrames representing
    the components of a graph network

    Parameters
    ----------
    transit_nodes : pandas.DataFrame
    transit_edges : pandas.DataFrame
    net_connector_edges : pandas.DataFrame
    osm_nodes : pandas.DataFrame
    osm_edges : pandas.DataFrame
    net_nodes : pandas.DataFrame
    net_edges : pandas.DataFrame
    """

    def __init__(self,
                 transit_nodes=pd.DataFrame(),
                 transit_edges=pd.DataFrame(),
                 net_connector_edges=pd.DataFrame(),
                 osm_nodes=pd.DataFrame(),
                 osm_edges=pd.DataFrame(),
                 net_nodes=pd.DataFrame(),
                 net_edges=pd.DataFrame()):
        self.transit_nodes = transit_nodes
        self.transit_edges = transit_edges
        self.net_connector_edges = net_connector_edges
        self.osm_nodes = osm_nodes
        self.osm_edges = osm_edges
        self.net_nodes = net_nodes
        self.net_edges = net_edges


# instantiate the UrbanAccess network object
ua_network = urbanaccess_network()


def _nearest_neighbor(df1, df2):
    """
    For a DataFrame of xy coordinates find the nearest xy
    coordinates in a subsequent DataFrame

    Parameters
    ----------
    df1 : pandas.DataFrame
        DataFrame of records to return as the nearest record to records in df2
    df2 : pandas.DataFrame
        DataFrame of records with xy coordinates for which to find the
        nearest record in df1 for
    Returns
    -------
    df1.index.values[indexes] : pandas.Series
        index of records in df1 that are nearest to the coordinates in df2
    """

    kdt = KDTree(df1.as_matrix())
    indexes = kdt.query(df2.as_matrix(), k=1, return_distance=False)
    return df1.index.values[indexes]


def integrate_network(urbanaccess_network, headways=False,
                      urbanaccess_gtfsfeeds_df=None, headway_statistic='mean'):
    """
    Create an integrated network comprised of transit and OSM nodes and edges
    by connecting the transit network with the osm network.
    travel time is in units of minutes

    Parameters
    ----------
    urbanaccess_network : object
        ua_network object with transit_edges, transit_nodes,
        osm_edges, osm_nodes
    headways : bool, optional
        if true, route stop level headways calculated in a previous step
        will be applied to the osm to transit connector
        edge travel time weights as an approximate measure
        of average passenger transit stop waiting time.
    urbanaccess_gtfsfeeds_df : object, optional
        required if headways is true; the gtfsfeeds_dfs object that holds
        the corresponding headways and stops DataFrames
    headway_statistic : {'mean', 'std', 'min', 'max'}, optional
        required if headways is true; route stop headway
        statistic to apply to the osm to transit connector edges:
        mean, std, min, max. Default is mean.

    Returns
    -------
    urbanaccess_network : object
    urbanaccess_network.transit_edges : pandas.DataFrame
    urbanaccess_network.transit_nodes : pandas.DataFrame
    urbanaccess_network.osm_edges : pandas.DataFrame
    urbanaccess_network.osm_nodes : pandas.DataFrame
    urbanaccess_network.net_connector_edges : pandas.DataFrame
    urbanaccess_network.net_edges : pandas.DataFrame
    urbanaccess_network.net_nodes : pandas.DataFrame
    """

    if urbanaccess_network is None:
        raise ValueError('urbanaccess_network is not specified')
    if urbanaccess_network.transit_edges.empty \
            or urbanaccess_network.transit_nodes.empty \
            or urbanaccess_network.osm_edges.empty \
            or urbanaccess_network.osm_nodes.empty:
        raise ValueError(
            'one of the network objects: transit_edges, transit_nodes, '
            'osm_edges, or osm_nodes were found to be empty.')

    log('Loaded UrbanAccess network components comprised of:')
    log('     Transit: {:,} nodes and {:,} edges;'.format(
        len(urbanaccess_network.transit_nodes),
        len(urbanaccess_network.transit_edges)))
    log('     OSM: {:,} nodes and {:,} edges'.format(
        len(urbanaccess_network.osm_nodes),
        len(urbanaccess_network.osm_edges)))

    if not isinstance(headways, bool):
        raise ValueError('headways must be bool type')

    if headways:

        if urbanaccess_gtfsfeeds_df is None or \
                urbanaccess_gtfsfeeds_df.headways.empty or \
                urbanaccess_gtfsfeeds_df.stops.empty:
            raise ValueError(
                'stops and headway DataFrames were not found in the '
                'urbanaccess_gtfsfeeds object. Please create these '
                'DataFrames in order to use headways.')

        valid_stats = ['mean', 'std', 'min', 'max']
        if headway_statistic not in valid_stats or not isinstance(
                headway_statistic, str):
            raise ValueError('{} is not a supported statistic or is not a '
                             'string'.format(headway_statistic))

        transit_edge_cols = urbanaccess_network.transit_edges.columns
        if 'node_id_from' not in transit_edge_cols or 'from' in \
                transit_edge_cols:
            urbanaccess_network.transit_edges.rename(
                columns={'from': 'node_id_from'}, inplace=True)
        if 'node_id_to' not in transit_edge_cols or 'to' in transit_edge_cols:
            urbanaccess_network.transit_edges.rename(
                columns={'to': 'node_id_to'}, inplace=True)

        urbanaccess_network.transit_edges['node_id_route_from'] = (
            urbanaccess_network.transit_edges['node_id_from'].str.cat(
                urbanaccess_network.transit_edges['unique_route_id'].astype(
                    'str'), sep='_'))
        urbanaccess_network.transit_edges['node_id_route_to'] = (
            urbanaccess_network.transit_edges['node_id_to'].str.cat(
                urbanaccess_network.transit_edges['unique_route_id'].astype(
                    'str'), sep='_'))

        urbanaccess_network.transit_nodes = _route_id_to_node(
            stops_df=urbanaccess_gtfsfeeds_df.stops,
            edges_w_routes=urbanaccess_network.transit_edges)

        net_connector_edges = _connector_edges(
            osm_nodes=urbanaccess_network.osm_nodes,
            transit_nodes=urbanaccess_network.transit_nodes,
            travel_speed_mph=3)

        urbanaccess_network.net_connector_edges = _add_headway_impedance(
            ped_to_transit_edges_df=net_connector_edges,
            headways_df=urbanaccess_gtfsfeeds_df.headways,
            headway_statistic=headway_statistic)

    else:
        urbanaccess_network.net_connector_edges = _connector_edges(
            osm_nodes=urbanaccess_network.osm_nodes,
            transit_nodes=urbanaccess_network.transit_nodes,
            travel_speed_mph=3)

    # change cols in transit edges and nodes
    if headways:
        urbanaccess_network.transit_edges.rename(columns={
            'node_id_route_from': 'from', 'node_id_route_to': 'to'},
            inplace=True)
        urbanaccess_network.transit_edges.drop(['node_id_from', 'node_id_to'],
                                               inplace=True, axis=1)
        urbanaccess_network.transit_nodes.reset_index(inplace=True, drop=False)
        urbanaccess_network.transit_nodes.rename(
            columns={'node_id_route': 'id'}, inplace=True)
    else:
        urbanaccess_network.transit_edges.rename(
            columns={'node_id_from': 'from', 'node_id_to': 'to'}, inplace=True)
        urbanaccess_network.transit_nodes.reset_index(inplace=True, drop=False)
        urbanaccess_network.transit_nodes.rename(columns={'node_id': 'id'},
                                                 inplace=True)

    # concat all network components
    urbanaccess_network.net_edges = pd.concat(
        [urbanaccess_network.transit_edges,
         urbanaccess_network.osm_edges,
         urbanaccess_network.net_connector_edges], axis=0)

    urbanaccess_network.net_nodes = pd.concat(
        [urbanaccess_network.transit_nodes,
         urbanaccess_network.osm_nodes], axis=0)

    urbanaccess_network.net_edges, urbanaccess_network.net_nodes = \
        _format_pandana_edges_nodes(edge_df=urbanaccess_network.net_edges,
                                    node_df=urbanaccess_network.net_nodes)

    success_msg_1 = ('Network edge and node network integration completed '
                     'successfully resulting in a total of {:,} nodes '
                     'and {:,} edges:')
    success_msg_2 = '     Transit: {:,} nodes {:,} edges;'
    success_msg_3 = '     OSM: {:,} nodes {:,} edges; and'
    success_msg_4 = '     OSM/Transit connector: {:,} edges.'
    log(success_msg_1.format(len(urbanaccess_network.net_nodes),
                             len(urbanaccess_network.net_edges)))
    log(success_msg_2.format(len(urbanaccess_network.transit_nodes),
                             len(urbanaccess_network.transit_edges)))
    log(success_msg_3.format(len(urbanaccess_network.osm_nodes),
                             len(urbanaccess_network.osm_edges)))
    log(success_msg_4.format(len(urbanaccess_network.net_connector_edges)))

    return urbanaccess_network


def _add_headway_impedance(ped_to_transit_edges_df, headways_df,
                           headway_statistic='mean'):
    """
    Add route stop level headways to the osm to transit connector
    travel time weight column

    Parameters
    ----------
    ped_to_transit_edges_df : pandas.DataFrame
        DataFrame of the osm to transit connectors
    headways_df : pandas.DataFrame
        headways DataFrame
    headway_statistic : {'mean', 'std', 'min', 'max'}, optional
        required if headways is true; route stop headway statistic to apply
        to the osm to transit connector edges:
        mean, std, min, max. Default is mean.

    Returns
    -------
    osm_to_transit_wheadway : pandas.DataFrame

    """

    start_time = time.time()

    log(
        '{} route stop headway will be used for pedestrian to transit edge '
        'impedance.'.format(
            headway_statistic))

    osm_to_transit_wheadway = pd.merge(ped_to_transit_edges_df, headways_df[
        [headway_statistic, 'node_id_route']],
                                       how='left', left_on=['to'],
                                       right_on=['node_id_route'], sort=False,
                                       copy=False)
    osm_to_transit_wheadway['weight_tmp'] = osm_to_transit_wheadway[
                                                'weight'] + (
                                            osm_to_transit_wheadway[
                                                headway_statistic] / 2.0)
    osm_to_transit_wheadway['weight_tmp'].fillna(
        osm_to_transit_wheadway['weight'], inplace=True)
    osm_to_transit_wheadway.drop('weight', axis=1, inplace=True)
    osm_to_transit_wheadway.rename(columns={'weight_tmp': 'weight'},
                                   inplace=True)

    log('Headway impedance calculation completed. Took {:,.2f} seconds'.format(
        time.time() - start_time))

    return osm_to_transit_wheadway


def _route_id_to_node(stops_df, edges_w_routes):
    """
    Assign route ids to the transit nodes table

    Parameters
    ----------
    stops_df : pandas.DataFrame
        processed gtfs stops DataFrame
    edges_w_routes : pandas.DataFrame
        transit edge DataFrame that has route id information

    Returns
    -------
    transit_nodes_wroutes : pandas.DataFrame

    """
    start_time = time.time()

    # create unique stop ids
    stops_df['unique_stop_id'] = (
        stops_df['stop_id'].str.cat(
            stops_df['unique_agency_id'].astype('str'), sep='_'))

    tmp1 = pd.merge(edges_w_routes[['node_id_from',
                                    'node_id_route_from']],
                    stops_df[['unique_stop_id',
                              'stop_lat',
                              'stop_lon']],
                    how='left', left_on='node_id_from',
                    right_on='unique_stop_id', sort=False, copy=False)
    tmp1.rename(columns={'node_id_route_from': 'node_id_route',
                         'stop_lon': 'x',
                         'stop_lat': 'y'},
                inplace=True)
    tmp2 = pd.merge(edges_w_routes[['node_id_to',
                                    'node_id_route_to']],
                    stops_df[['unique_stop_id',
                              'stop_lat',
                              'stop_lon']],
                    how='left',
                    left_on='node_id_to',
                    right_on='unique_stop_id', sort=False, copy=False)
    tmp2.rename(columns={'node_id_route_to': 'node_id_route',
                         'stop_lon': 'x',
                         'stop_lat': 'y'},
                inplace=True)

    transit_nodes_wroutes = pd.concat([tmp1[['node_id_route', 'x', 'y']],
                                       tmp2[['node_id_route', 'x', 'y']]],
                                      axis=0)

    transit_nodes_wroutes.drop_duplicates(subset='node_id_route',
                                          keep='first',
                                          inplace=True)
    # set node index to be unique stop id
    transit_nodes_wroutes = transit_nodes_wroutes.set_index('node_id_route')

    log(
        'routes successfully joined to transit nodes. Took {:,'
        '.2f} seconds'.format(
            time.time() - start_time))

    return transit_nodes_wroutes


def _connector_edges(osm_nodes, transit_nodes, travel_speed_mph=3):
    """
    Generate the connector edges between the osm and transit edges and
    weight by travel time

    Parameters
    ----------
    osm_nodes : pandas.DataFrame
        osm nodes DataFrame
    transit_nodes : pandas.DataFrame
        transit nodes DataFrame
    travel_speed_mph : int, optional
        travel speed to use to calculate travel time across a
        distance on a edge. units are in miles per hour (MPH)
        for pedestrian travel this is assumed to be 3 MPH

    Returns
    -------
    net_connector_edges : pandas.DataFrame

    """
    start_time = time.time()

    transit_nodes['nearest_osm_node'] = _nearest_neighbor(
        osm_nodes[['x', 'y']],
        transit_nodes[['x', 'y']])

    net_connector_edges = []

    for transit_node_id, row in transit_nodes.iterrows():
        # create new edge between the node in df2 (transit)
        # and the node in openstreetmap (pedestrian)

        osm_node_id = int(row['nearest_osm_node'])
        osm_row = osm_nodes.loc[osm_node_id]

        distance = vincenty((row['y'], row['x']),
                            (osm_row['y'], osm_row['x'])).miles
        time_ped_to_transit = distance / travel_speed_mph * 60
        time_transit_to_ped = distance / travel_speed_mph * 60

        # save the edge
        net_type = 'transit to osm'
        net_connector_edges.append((transit_node_id, osm_node_id,
                                    time_transit_to_ped, net_type))
        # make the edge bi-directional
        net_type = 'osm to transit'
        net_connector_edges.append((osm_node_id, transit_node_id,
                                    time_ped_to_transit, net_type))

    net_connector_edges = pd.DataFrame(net_connector_edges,
                                       columns=["from", "to",
                                                "weight", "net_type"])

    log(
        'Connector edges between the OSM and transit network nodes '
        'successfully completed. Took {:,.2f} seconds'.format(
            time.time() - start_time))

    return net_connector_edges


def _format_pandana_edges_nodes(edge_df, node_df):
    """
    Perform final formatting on nodes and edge DataFrames to prepare them
    for use in Pandana.
    Formatting mainly consists of creating a unique node id and edge from
    and to id that is an integer
    per Pandana requirements.

    Parameters
    ----------
    edge_df : pandas.DataFrame
        integrated transit and osm edge DataFrame
    node_df : pandas.DataFrame
        integrated transit and osm node DataFrame

    Returns
    -------
    edge_df_wnumericid, node_df : pandas.DataFrame

    """
    start_time = time.time()

    # pandana requires ids that are integer: for nodes - make it the index,
    # for edges make it the from and to columns
    node_df['id_int'] = range(1, len(node_df) + 1)

    edge_df.rename(columns={'id': 'edge_id'}, inplace=True)
    tmp = pd.merge(edge_df, node_df[['id', 'id_int']], left_on='from',
                   right_on='id', sort=False, copy=False, how='left')
    tmp['from_int'] = tmp['id_int']
    tmp.drop(['id_int', 'id'], axis=1, inplace=True)
    edge_df_wnumericid = pd.merge(tmp, node_df[['id', 'id_int']], left_on='to',
                                  right_on='id', sort=False, copy=False,
                                  how='left')
    edge_df_wnumericid['to_int'] = edge_df_wnumericid['id_int']
    edge_df_wnumericid.drop(['id_int', 'id'], axis=1, inplace=True)
    # turn mixed dtype cols into all same format
    col_list = edge_df_wnumericid.select_dtypes(include=['object']).columns
    for col in col_list:
        try:
            edge_df_wnumericid[col] = edge_df_wnumericid[col].astype(str)
        # deal with edge cases where typically the name of a street is not
        # in a uniform string encoding such as names with accents
        except UnicodeEncodeError:
            log('Fixed unicode error in {} column'.format(col))
            edge_df_wnumericid[col] = edge_df_wnumericid[col].str.encode(
                'utf-8')

    node_df.set_index('id_int', drop=True, inplace=True)
    # turn mixed dtype col into all same format
    node_df['id'] = node_df['id'].astype(str)
    if 'nearest_osm_node' in node_df.columns:
        node_df.drop(['nearest_osm_node'], axis=1, inplace=True)

    log('Edge and node tables formatted for Pandana with integer node ids: '
        'id_int, to_int, and from_int. Took {:,.2f} seconds'.format(
            time.time() - start_time))
    return edge_df_wnumericid, node_df


def save_network(urbanaccess_network, filename,
                 dir=config.settings.data_folder,
                 overwrite_key=False, overwrite_hdf5=False):
    """
    Write a urbanaccess_network integrated nodes and edges to a node and edge
    table in a hdf5 file

    Parameters
    ----------
    urbanaccess_network : object
        urbanaccess_network object with net_edges and net_nodes DataFrames
    filename : string
        name of the hdf5 file to save with .h5 extension
    dir : string, optional
        directory to save hdf5 file
    overwrite_key : bool, optional
        if true any existing table with the specified key name will be
        overwritten
    overwrite_hdf5 : bool, optional
        if true any existing hdf5 file with the specified name in the
        specified directory will be overwritten

    Returns
    -------
    None
    """
    if urbanaccess_network is None or urbanaccess_network.net_edges.empty or \
            urbanaccess_network.net_nodes.empty:
        raise ValueError('Either no urbanaccess_network specified or '
                         'net_edges or net_nodes are empty.')

    df_to_hdf5(data=urbanaccess_network.net_edges, key='edges',
               overwrite_key=overwrite_key, dir=dir,
               filename=filename, overwrite_hdf5=overwrite_hdf5)
    df_to_hdf5(data=urbanaccess_network.net_nodes, key='nodes',
               overwrite_key=overwrite_key, dir=dir, filename=filename,
               overwrite_hdf5=overwrite_hdf5)


def load_network(dir=config.settings.data_folder, filename=None):
    """
    Read an integrated network node and edge data from a hdf5 file to
    a urbanaccess_network object

    Parameters
    ----------
    dir : string, optional
        directory to read hdf5 file
    filename : string
        name of the hdf5 file to read with .h5 extension

    Returns
    -------
    ua_network : object
        urbanaccess_network object with net_edges and net_nodes DataFrames
    ua_network.net_edges : object
    ua_network.net_nodes : object
    """
    ua_network.net_edges = hdf5_to_df(dir=dir, filename=filename, key='edges')
    ua_network.net_nodes = hdf5_to_df(dir=dir, filename=filename, key='nodes')

    return ua_network



import pandas as pd


# TODO: make class CamelCase
class urbanaccess_gtfs_df(object):
    """
    A collection of dataframes representing the standardized and
    merged metropolitan-wide transit network from multiple
    GTFS feeds.

    Parameters
    ----------
    gtfsfeeds_dfs : object
        processed dataframes of corresponding GTFS feed text files
    gtfsfeeds_dfs.stops : pandas.DataFrame
    gtfsfeeds_dfs.routes : pandas.DataFrame
    gtfsfeeds_dfs.trips : pandas.DataFrame
    gtfsfeeds_dfs.stop_times : pandas.DataFrame
    gtfsfeeds_dfs.calendar : pandas.DataFrame
    gtfsfeeds_dfs.calendar_dates : pandas.DataFrame
    gtfsfeeds_dfs.stop_times_int : pandas.DataFrame
    gtfsfeeds_dfs.headways : pandas.DataFrame
    """

    def __init__(self,
                 stops=pd.DataFrame(),
                 routes=pd.DataFrame(),
                 trips=pd.DataFrame(),
                 stop_times=pd.DataFrame(),
                 calendar=pd.DataFrame(),
                 calendar_dates=pd.DataFrame(),
                 stop_times_int=pd.DataFrame(),
                 headways=pd.DataFrame()):
        self.stops = stops
        self.routes = routes
        self.trips = trips
        self.stop_times = stop_times
        self.calendar = calendar
        self.calendar_dates = calendar_dates
        self.stop_times_int = stop_times_int
        self.headways = headways


# instantiate the UrbanAccess gtfs feed dataframe object
gtfsfeeds_dfs = urbanaccess_gtfs_df()



# Install setuptools if not installed.
try:
    import setuptools
except ImportError:
    from ez_setup import use_setuptools

    use_setuptools()

from setuptools import setup, find_packages


# read README as the long description
with open('README.rst', 'r') as f:
    long_description = f.read()

description = 'A tool for creating GTFS transit and OSM pedestrian networks ' \
              'for use in Pandana accessibility analyses.'

setup(
    name='urbanaccess',
    version='0.2.0',
    license='AGPL',
    description=description,
    long_description=long_description,
    author='UrbanSim Inc. and Samuel D. Blanchard',
    url='https://github.com/UDST/urbanaccess',
    classifiers=[
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: GNU Affero General Public License v3'
    ],
    packages=find_packages(exclude=['*.tests']),
    install_requires=[
        'requests >= 2.9.1',
        'six >= 1.11',
        'pandas >= 0.17.0',
        'numpy >= 1.11',
        'osmnet >= 0.1.4',
        'pandana >= 0.2.0',
        'matplotlib >= 2.0',
        'geopy >= 1.11.0',
        'pyyaml >= 3.11',
        'scikit-learn >= 0.17.1'
    ]
)



import os
import codecs
import re
import time
import pandas as pd
import six

from urbanaccess import config
from urbanaccess.utils import log
from urbanaccess.gtfs.gtfsfeeds_dataframe import gtfsfeeds_dfs
from urbanaccess.gtfs import utils_validation
from urbanaccess.gtfs import utils_format


def _standardize_txt(csv_rootpath=os.path.join(config.settings.data_folder,
                                               'gtfsfeed_text')):
    """
    Standardize all text files inside a GTFS feed for machine readability

    Parameters
    ----------
    csv_rootpath : str, optional
        root path where all gtfs feeds that make up a contiguous metropolitan
        area are stored

    Returns
    -------
    None
    """

    gtfsfiles_to_use = ['stops.txt', 'routes.txt', 'trips.txt',
                        'stop_times.txt', 'calendar.txt',
                        'agency.txt', 'calendar_dates.txt']

    if six.PY2:
        _txt_encoder_check(gtfsfiles_to_use, csv_rootpath)
    _txt_header_whitespace_check(gtfsfiles_to_use, csv_rootpath)


def _txt_encoder_check(gtfsfiles_to_use,
                       csv_rootpath=os.path.join(
                           config.settings.data_folder,
                           'gtfsfeed_text')):
    """
    Standardize all text files inside a GTFS feed for encoding problems.
    Has not been updated for Python 3.

    Parameters
    ----------
    gtfsfiles_to_use : list
        list of gtfs feed txt files to utilize
    csv_rootpath : str, optional
        root path where all gtfs feeds that make up a contiguous metropolitan
        area are stored

    Returns
    -------
    None
    """
    # UnicodeDecodeError
    start_time = time.time()

    folderlist = [foldername for foldername in os.listdir(csv_rootpath) if
                  os.path.isdir(os.path.join(csv_rootpath, foldername))]

    if not folderlist:
        folderlist = [csv_rootpath]

    for folder in folderlist:
        textfilelist = [textfilename for textfilename in
                        os.listdir(os.path.join(csv_rootpath, folder)) if
                        textfilename.endswith(".txt")]

        for textfile in textfilelist:
            if textfile in gtfsfiles_to_use:
                # Read from file
                file_open = open(os.path.join(csv_rootpath, folder, textfile))
                raw = file_open.read()
                file_open.close()
                if raw.startswith(codecs.BOM_UTF8):
                    raw = raw.replace(codecs.BOM_UTF8, '', 1)
                    # Write to file
                    file_open = open(
                        os.path.join(csv_rootpath, folder, textfile), 'w')
                    file_open.write(raw)
                    file_open.close()

    log('GTFS text file encoding check completed. Took {:,.2f} seconds'.format(
        time.time() - start_time))


def _txt_header_whitespace_check(gtfsfiles_to_use,
                                 csv_rootpath=os.path.join(
                                     config.settings.data_folder,
                                     'gtfsfeed_text')):
    """
    Standardize all text files inside a GTFS feed to remove whitespace
    in headers

    Parameters
    ----------
    gtfsfiles_to_use : list
        list of gtfs feed txt files to utilize
    csv_rootpath : str, optional
        root path where all gtfs feeds that make up a contiguous metropolitan
        area are stored

    Returns
    -------
    None
    """
    start_time = time.time()

    folderlist = [foldername for foldername in os.listdir(csv_rootpath) if
                  os.path.isdir(os.path.join(csv_rootpath, foldername))]

    if not folderlist:
        folderlist = [csv_rootpath]

    for folder in folderlist:
        textfilelist = [textfilename for textfilename in
                        os.listdir(os.path.join(csv_rootpath, folder)) if
                        textfilename.endswith(".txt")]

        for textfile in textfilelist:
            if textfile in gtfsfiles_to_use:
                # Read from file
                with open(os.path.join(csv_rootpath, folder, textfile)) as f:
                    lines = f.readlines()
                lines[0] = re.sub(r'\s+', '', lines[0]) + '\n'
                # Write to file
                try:
                    with open(os.path.join(csv_rootpath, folder, textfile),
                              'w') as f:
                        f.writelines(lines)
                except Exception:
                    log('Unable to read {}. Check that file is not currently'
                        'being read or is not already in memory as this is '
                        'likely the cause of the error.'
                        ''.format(os.path.join(csv_rootpath,
                                               folder, textfile)))
    log(
        'GTFS text file header whitespace check completed. Took {:,'
        '.2f} seconds'.format(
            time.time() - start_time))


def gtfsfeed_to_df(gtfsfeed_path=None, validation=False, verbose=True,
                   bbox=None, remove_stops_outsidebbox=None,
                   append_definitions=False):
    """
    Read all GTFS feed components as a dataframe in a gtfsfeeds_dfs object and
    merge all individual GTFS feeds into a regional metropolitan data table.
    Optionally, data can also be validated before its use.

    Parameters
    ----------
    gtfsfeed_path : str, optional
        root path where all gtfs feeds that make up a contiguous metropolitan
        area are stored
    validation : bool
        if true, the validation check on stops checking for stops outside
        of a bounding box and stop coordinate
        hemisphere will be run. this is required to remove stops outside of
        a bbox
    verbose : bool
        if true and stops are found outside of the bbox, the stops that are
        outside will be printed for your reference
    bbox : tuple
        Bounding box formatted as a 4 element tuple:
        (lng_max, lat_min, lng_min, lat_max)
        example: (-122.304611,37.798933,-122.263412,37.822802)
        a bbox can be extracted for an area using: the CSV format bbox
        from http://boundingbox.klokantech.com/
    remove_stops_outsidebbox : bool
        if true stops that are outside the bbox will be removed
    append_definitions : bool
        if true, columns that use the GTFS data schema for their attribute
        codes will have the corresponding GTFS definition information of
        that code appended to the resulting dataframes for reference

    Returns
    -------
    gtfsfeeds_dfs : object
        processed dataframes of corresponding GTFS feed text files
    gtfsfeeds_dfs.stops : pandas.DataFrame
    gtfsfeeds_dfs.routes : pandas.DataFrame
    gtfsfeeds_dfs.trips : pandas.DataFrame
    gtfsfeeds_dfs.stop_times : pandas.DataFrame
    gtfsfeeds_dfs.calendar : pandas.DataFrame
    gtfsfeeds_dfs.calendar_dates : pandas.DataFrame
    """

    merged_stops_df = pd.DataFrame()
    merged_routes_df = pd.DataFrame()
    merged_trips_df = pd.DataFrame()
    merged_stop_times_df = pd.DataFrame()
    merged_calendar_df = pd.DataFrame()
    merged_calendar_dates_df = pd.DataFrame()

    start_time = time.time()

    if gtfsfeed_path is None:
        gtfsfeed_path = os.path.join(config.settings.data_folder,
                                     'gtfsfeed_text')
        if not os.path.exists(gtfsfeed_path):
            raise ValueError('{} does not exist'.format(gtfsfeed_path))
    else:
        if not os.path.exists(gtfsfeed_path):
            raise ValueError('{} does not exist'.format(gtfsfeed_path))
    if not isinstance(gtfsfeed_path, str):
        raise ValueError('gtfsfeed_path must be a string')

    if validation:
        if bbox is None or remove_stops_outsidebbox is None or verbose is \
                None:
            raise ValueError(
                'Attempted to run validation but bbox, verbose, and or '
                'remove_stops_outsidebbox were set to None. These parameters '
                'must be specified for validation.')

    _standardize_txt(csv_rootpath=gtfsfeed_path)

    folderlist = [foldername for foldername in os.listdir(gtfsfeed_path) if
                  os.path.isdir(os.path.join(gtfsfeed_path, foldername))]
    if not folderlist:
        folderlist = [gtfsfeed_path]

    for index, folder in enumerate(folderlist):

        # print break to visually separate each gtfs feed log
        log('--------------------------------')
        log('Processing GTFS feed: {!s}'.format(os.path.split(folder)[1]))

        textfilelist = [textfilename for textfilename in
                        os.listdir(os.path.join(gtfsfeed_path, folder)) if
                        textfilename.endswith(".txt")]
        required_gtfsfiles = ['stops.txt', 'routes.txt', 'trips.txt',
                              'stop_times.txt', 'calendar.txt']
        optional_gtfsfiles = ['agency.txt', 'calendar_dates.txt']
        for required_file in required_gtfsfiles:
            if required_file not in textfilelist:
                raise ValueError(
                    '{} is a required GTFS text file and was not found in '
                    'folder {}'.format(
                        required_file,
                        os.path.join(gtfsfeed_path, folder)))

        for textfile in required_gtfsfiles:
            if textfile == 'stops.txt':
                stops_df = utils_format._read_gtfs_stops(
                    textfile_path=os.path.join(gtfsfeed_path, folder),
                    textfile=textfile)
            if textfile == 'routes.txt':
                routes_df = utils_format._read_gtfs_routes(
                    textfile_path=os.path.join(gtfsfeed_path, folder),
                    textfile=textfile)
            if textfile == 'trips.txt':
                trips_df = utils_format._read_gtfs_trips(
                    textfile_path=os.path.join(gtfsfeed_path, folder),
                    textfile=textfile)
            if textfile == 'stop_times.txt':
                stop_times_df = utils_format._read_gtfs_stop_times(
                    textfile_path=os.path.join(gtfsfeed_path, folder),
                    textfile=textfile)
            if textfile == 'calendar.txt':
                calendar_df = utils_format._read_gtfs_calendar(
                    textfile_path=os.path.join(gtfsfeed_path, folder),
                    textfile=textfile)

        for textfile in optional_gtfsfiles:
            if textfile == 'agency.txt':
                if textfile in textfilelist:
                    agency_df = utils_format._read_gtfs_agency(
                        textfile_path=os.path.join(gtfsfeed_path, folder),
                        textfile=textfile)
                else:
                    agency_df = pd.DataFrame()
            if textfile == 'calendar_dates.txt':
                if textfile in textfilelist:
                    calendar_dates_df = utils_format._read_gtfs_calendar_dates(
                        textfile_path=os.path.join(gtfsfeed_path, folder),
                        textfile=textfile)
                else:
                    calendar_dates_df = pd.DataFrame()

        stops_df, routes_df, trips_df, stop_times_df, calendar_df, \
            calendar_dates_df = (utils_format
                                 ._add_unique_agencyid(
                                    agency_df=agency_df,
                                    stops_df=stops_df,
                                    routes_df=routes_df,
                                    trips_df=trips_df,
                                    stop_times_df=stop_times_df,
                                    calendar_df=calendar_df,
                                    calendar_dates_df=calendar_dates_df,
                                    nulls_as_folder=True,
                                    feed_folder=os.path.join(
                                        gtfsfeed_path, folder)))

        stops_df, routes_df, trips_df, stop_times_df, calendar_df, \
            calendar_dates_df = (utils_format
                                 ._add_unique_gtfsfeed_id(
                                    stops_df=stops_df,
                                    routes_df=routes_df,
                                    trips_df=trips_df,
                                    stop_times_df=stop_times_df,
                                    calendar_df=calendar_df,
                                    calendar_dates_df=calendar_dates_df,
                                    feed_folder=folder,
                                    feed_number=index+1))

        if validation:
            stops_df = (utils_validation
                        ._validate_gtfs(
                            stops_df=stops_df,
                            feed_folder=os.path.join(gtfsfeed_path, folder),
                            verbose=verbose,
                            bbox=bbox,
                            remove_stops_outsidebbox=remove_stops_outsidebbox))
            if remove_stops_outsidebbox:
                stops_inside_bbox = list(stops_df['stop_id'])
                stop_times_df = stop_times_df[stop_times_df['stop_id'].isin(
                    stops_inside_bbox)]

        stops_df = (utils_format
                    ._append_route_type(
                        stops_df=stops_df,
                        stop_times_df=stop_times_df,
                        routes_df=routes_df[['route_id', 'route_type']],
                        trips_df=trips_df[['trip_id', 'route_id']],
                        info_to_append='route_type_to_stops'))
        stop_times_df = (utils_format
                         ._append_route_type(
                            stops_df=stops_df,
                            stop_times_df=stop_times_df,
                            routes_df=routes_df[['route_id', 'route_type']],
                            trips_df=trips_df[['trip_id', 'route_id']],
                            info_to_append='route_type_to_stop_times'))

        merged_stops_df = merged_stops_df.append(stops_df, ignore_index=True)
        merged_routes_df = merged_routes_df.append(routes_df,
                                                   ignore_index=True)
        merged_trips_df = merged_trips_df.append(trips_df, ignore_index=True)
        merged_stop_times_df = merged_stop_times_df.append(stop_times_df,
                                                           ignore_index=True)
        merged_calendar_df = merged_calendar_df.append(calendar_df,
                                                       ignore_index=True)
        merged_calendar_dates_df = merged_calendar_dates_df.append(
            calendar_dates_df, ignore_index=True)

        # print break to visually separate each gtfs feed log
        log('--------------------------------')

    if append_definitions:
        merged_stops_df, merged_routes_df, merged_stop_times_df, \
            merged_trips_df = (utils_format
                               ._add_txt_definitions(
                                    stops_df=merged_stops_df,
                                    routes_df=merged_routes_df,
                                    stop_times_df=merged_stop_times_df,
                                    trips_df=merged_trips_df))

    merged_stop_times_df = utils_format._timetoseconds(
        df=merged_stop_times_df, time_cols=['departure_time'])

    # set gtfsfeeds_dfs object to merged GTFS dfs
    gtfsfeeds_dfs.stops = merged_stops_df
    gtfsfeeds_dfs.routes = merged_routes_df
    gtfsfeeds_dfs.trips = merged_trips_df
    gtfsfeeds_dfs.stop_times = merged_stop_times_df
    gtfsfeeds_dfs.calendar = merged_calendar_df
    gtfsfeeds_dfs.calendar_dates = merged_calendar_dates_df

    # TODO: add to print the list of gtfs feed txt files read in for each feed
    log('{:,} GTFS feed file(s) successfully read as dataframes:'.format(
        len(folderlist)))
    for folder in folderlist:
        log('     {}'.format(os.path.split(folder)[1]))
    log('     Took {:,.2f} seconds'.format(time.time() - start_time))

    return gtfsfeeds_dfs
